import sqlite3
import warnings
from dataclasses import dataclass, field, asdict
from tempfile import NamedTemporaryFile
from typing import (
    Iterable,
    Dict,
    Optional,
    TYPE_CHECKING,
    Union,
    List,
    Tuple,
)

from docarray.array.storage.sqlite.helper import initialize_table
from docarray.array.storage.base.backend import BaseBackendMixin
from docarray.helper import random_identity, dataclass_from_dict

if TYPE_CHECKING:
    from docarray.typing import DocumentArraySourceType


def _sanitize_table_name(table_name: str, raise_warning=True) -> str:
    ret = ''.join(c for c in table_name if c.isalnum() or c == '_')
    if ret != table_name and raise_warning:
        warnings.warn(f'The table name is changed to {ret} due to illegal characters')
    return ret


@dataclass
class SqliteConfig:
    connection: Optional[Union[str, 'sqlite3.Connection']] = None
    table_name: Optional[str] = None
    serialize_config: Dict = field(default_factory=dict)
    conn_config: Dict = field(default_factory=dict)
    journal_mode: str = 'DELETE'
    synchronous: str = 'OFF'


class BackendMixin(BaseBackendMixin):
    """Provide necessary functions to enable this storage backend."""

    schema_version = '0'

    def _sql(self, *args, **kwargs) -> 'sqlite3.Cursor':
        return self._cursor.execute(*args, **kwargs)

    def _commit(self):
        self._connection.commit()

    @property
    def _cursor(self) -> 'sqlite3.Cursor':
        return self._connection.cursor()

    def _init_storage(
        self,
        _docs: Optional['DocumentArraySourceType'] = None,
        config: Optional[Union[SqliteConfig, Dict]] = None,
        **kwargs,
    ):
        if not config:
            config = SqliteConfig()

        if isinstance(config, dict):
            config = dataclass_from_dict(SqliteConfig, config)

        from docarray import Document

        sqlite3.register_adapter(
            Document, lambda d: d.to_bytes(**config.serialize_config)
        )
        sqlite3.register_converter(
            'Document', lambda x: Document.from_bytes(x, **config.serialize_config)
        )

        _conn_kwargs = dict()
        _conn_kwargs.update(config.conn_config)
        if config.connection is None:
            config.connection = NamedTemporaryFile().name

        if isinstance(config.connection, str):
            self._connection = sqlite3.connect(
                config.connection,
                detect_types=sqlite3.PARSE_DECLTYPES,
                check_same_thread=False,
                **_conn_kwargs,
            )
        elif isinstance(config.connection, sqlite3.Connection):
            self._connection = config.connection
        else:
            raise TypeError(
                f'connection argument must be None or a string or a sqlite3.Connection, not `{type(config.connection)}`'
            )
        self._connection.execute(f'PRAGMA synchronous={config.synchronous}')
        self._connection.execute(f'PRAGMA journal_mode={config.journal_mode}')

        self._table_name = (
            _sanitize_table_name(self.__class__.__name__ + random_identity())
            if config.table_name is None
            else _sanitize_table_name(config.table_name)
        )
        self._persist = bool(config.table_name)
        config.table_name = self._table_name
        initialize_table(
            self._table_name, self.__class__.__name__, self.schema_version, self._cursor
        )
        self._connection.commit()
        self._config = config

        super()._init_storage()

        if _docs is None:
            return
        elif isinstance(_docs, Iterable):
            self.clear()
            self.extend(_docs)
        else:
            self.clear()
            if isinstance(_docs, Document):
                self.append(_docs)

    def _ensure_unique_config(
        self,
        config_root: dict,
        config_subindex: dict,
        config_joined: dict,
        subindex_name: str,
    ) -> dict:
        if 'table_name' not in config_subindex:
            subindex_table_name = _sanitize_table_name(
                config_joined['table_name'] + 'subindex' + subindex_name,
                raise_warning=False,
            )
            config_joined['table_name'] = subindex_table_name
        return config_joined

    def __getstate__(self):
        d = dict(self.__dict__)
        del d['_connection']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        _conn_kwargs = dict()
        _conn_kwargs.update(state['_config'].conn_config)
        self._connection = sqlite3.connect(
            state['_config'].connection,
            detect_types=sqlite3.PARSE_DECLTYPES,
            check_same_thread=False,
            **_conn_kwargs,
        )



from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from docarray.array.mixins import ParallelMixin, GroupMixin
from docarray.helper import protocol_and_compress_from_file_path

if TYPE_CHECKING:
    from docarray import Document, DocumentArray


class DocumentArrayLoader(ParallelMixin, GroupMixin):
    def __init__(
        self,
        path: Union[str, Path],
        protocol: str = 'protobuf',
        compress: Optional[str] = None,
        show_progress: bool = False,
    ):
        self._show_progress = show_progress
        self._filename = path
        self._protocol, self._compress = protocol_and_compress_from_file_path(
            path, protocol, compress
        )

        with open(path, 'rb') as f:
            version_numdocs_lendoc0 = f.read(9)
            # 8 bytes (uint64)
            self._len = int.from_bytes(
                version_numdocs_lendoc0[1:9], 'big', signed=False
            )
        self._iter = iter(self)

    def __iter__(self):
        from docarray import Document

        from docarray.array.mixins.io.pbar import get_progressbar
        from rich import filesize

        with open(self._filename, 'rb') as f:
            f.read(9)

            pbar, t = get_progressbar(
                'Deserializing', disable=not self._show_progress, total=self._len
            )

            with pbar:
                _total_size = 0
                pbar.start_task(t)
                for _ in range(self._len):
                    # 4 bytes (uint32)
                    len_current_doc_in_bytes = int.from_bytes(
                        f.read(4), 'big', signed=False
                    )
                    _total_size += len_current_doc_in_bytes
                    yield Document.from_bytes(
                        f.read(len_current_doc_in_bytes),
                        protocol=self._protocol,
                        compress=self._compress,
                    )
                    pbar.update(
                        t, advance=1, total_size=str(filesize.decimal(_total_size))
                    )

    def __len__(self):
        return self._len

    def __getitem__(self, item: list) -> 'DocumentArray':
        from docarray import DocumentArray

        da = DocumentArray()
        for _ in item:
            da.append(next(self._iter))
        return da



import base64
import typing

from docarray.dataclasses.types import Field, is_multimodal, _is_field
from docarray.dataclasses.types import AttributeType

if typing.TYPE_CHECKING:
    from docarray import Document, DocumentArray


class MultiModalMixin:
    @property
    def is_multimodal(self) -> bool:
        """
        Return true if this Document can be represented by a class wrapped
        by :meth:`docarray.dataclasses.types.dataclass`.
        """
        return 'multi_modal_schema' in self._metadata

    @classmethod
    def _from_dataclass(cls, obj) -> 'Document':
        if not is_multimodal(obj):
            raise TypeError(
                f'Object {type(obj).__name__} is not a `docarray.dataclass` instance'
            )

        from docarray import Document

        root = Document()
        tags = {}
        multi_modal_schema = {}
        for key, field in obj.__dataclass_fields__.items():
            attribute = getattr(obj, key)
            if attribute is None:
                continue

            if field.type in [str, int, float, bool] and not _is_field(field):
                tags[key] = attribute
                multi_modal_schema[key] = {
                    'attribute_type': AttributeType.PRIMITIVE,
                    'type': field.type.__name__,
                }

            elif field.type == bytes and not _is_field(field):
                tags[key] = base64.b64encode(attribute).decode()
                multi_modal_schema[key] = {
                    'attribute_type': AttributeType.PRIMITIVE,
                    'type': field.type.__name__,
                }
            elif isinstance(field.type, typing._GenericAlias):
                if field.type._name in ['List', 'Iterable']:
                    sub_type = field.type.__args__[0]
                    if sub_type in [str, int, float, bool]:
                        tags[key] = attribute
                        multi_modal_schema[key] = {
                            'attribute_type': AttributeType.ITERABLE_PRIMITIVE,
                            'type': f'{field.type._name}[{sub_type.__name__}]',
                        }

                    else:
                        chunk = Document()
                        for element in attribute:
                            doc, attribute_type = cls._from_obj(
                                element, sub_type, field
                            )
                            if attribute_type == AttributeType.DOCUMENT:
                                attribute_type = AttributeType.ITERABLE_DOCUMENT
                            elif attribute_type == AttributeType.NESTED:
                                attribute_type = AttributeType.ITERABLE_NESTED
                            else:
                                raise TypeError(
                                    f'Unsupported type annotation inside Iterable: {sub_type}'
                                )
                            chunk.chunks.append(doc)
                        multi_modal_schema[key] = {
                            'attribute_type': attribute_type,
                            'type': f'{field.type._name}[{sub_type.__name__}]',
                            'position': len(root.chunks),
                        }
                        root.chunks.append(chunk)
                else:
                    raise TypeError(
                        f'Unsupported type annotation on field `{field.type._name}`'
                    )
            else:
                doc, attribute_type = cls._from_obj(attribute, field.type, field)
                multi_modal_schema[key] = {
                    'attribute_type': attribute_type,
                    'type': field.type.__name__,
                    'position': len(root.chunks),
                }
                root.chunks.append(doc)

        # TODO: may have to modify this?
        root.tags = tags
        root._metadata['multi_modal_schema'] = multi_modal_schema

        return root

    def _get_mm_attr_postion(self, attr):
        if not self.is_multimodal:
            raise ValueError(
                'the Document does not correspond to a Multi Modal Document'
            )

        if attr not in self._metadata['multi_modal_schema']:
            raise ValueError(f'the Document schema does not contain attribute {attr}')

        pos = self._metadata['multi_modal_schema'][attr].get('position')
        if pos is None:
            raise ValueError(
                f'attribute {attr} is not a valid multi modal attribute.'
                f' One possible cause is the usage of a non-supported type in the dataclass definition.'
            )
        return int(pos)

    def get_multi_modal_attribute(self, attribute: str) -> 'DocumentArray':
        from docarray import DocumentArray

        position = self._get_mm_attr_postion(attribute)
        attribute_type = self._metadata['multi_modal_schema'][attribute][
            'attribute_type'
        ]

        if attribute_type in [AttributeType.DOCUMENT, AttributeType.NESTED]:
            return DocumentArray([self.chunks[position]])
        elif attribute_type in [
            AttributeType.ITERABLE_DOCUMENT,
            AttributeType.ITERABLE_NESTED,
        ]:
            return self.chunks[position].chunks
        else:
            raise ValueError(
                f'Invalid attribute {attribute}: must be a Document attribute or nested dataclass'
            )

    def set_multi_modal_attribute(
        self, attribute: str, value: typing.Union['Document', 'DocumentArray']
    ):
        position = self._get_mm_attr_postion(attribute)
        attribute_type = self._metadata['multi_modal_schema'][attribute][
            'attribute_type'
        ]

        if attribute_type in [AttributeType.DOCUMENT, AttributeType.NESTED]:
            self.chunks[position] = value
        elif attribute_type in [
            AttributeType.ITERABLE_DOCUMENT,
            AttributeType.ITERABLE_NESTED,
        ]:
            self.chunks[position].chunks = value
        else:
            raise ValueError(
                f'Invalid attribute {attribute}: must be a Document attribute or nested dataclass'
            )

    @classmethod
    def _from_obj(cls, obj, obj_type, field) -> typing.Tuple['Document', AttributeType]:
        attribute_type = AttributeType.DOCUMENT

        if is_multimodal(obj_type):
            doc = cls(obj)
            attribute_type = AttributeType.NESTED
        elif _is_field(field):
            doc = field.setter(obj)
        else:
            raise ValueError(f'Unsupported type annotation')
        return doc, attribute_type

    def _has_multimodal_attr(self, attr):
        try:
            data = super().__getattribute__('_data')
            has_data = bool(data)
        except AttributeError:
            return False

        has_metadata = has_data and getattr(self._data, '_metadata') is not None
        return (
            has_metadata
            and self.is_multimodal
            and attr in self._metadata['multi_modal_schema']
        )

    def __getattr__(self, attr):
        if self._has_multimodal_attr(attr):
            mm_attr_da = self.get_multi_modal_attribute(attr)
            return mm_attr_da if len(mm_attr_da) > 1 else mm_attr_da[0]
        else:
            raise AttributeError(f'{self.__class__.__name__} has no attribute {attr}')

    def __setattr__(self, attr, value):
        if self._has_multimodal_attr(attr):
            self.set_multi_modal_attribute(attr, value)
        else:
            object.__setattr__(self, attr, value)



# auto-generated from /Users/hanxiao/Documents/docarray/scripts/gen_doc_property_mixin.py
from typing import TYPE_CHECKING, Dict, List, Optional

if TYPE_CHECKING:
    from docarray.score import NamedScore
    from docarray.array.match import MatchArray
    from docarray.array.chunk import ChunkArray
    from docarray import DocumentArray
    from docarray.typing import ArrayType, StructValueType, DocumentContentType


class _PropertyMixin:
    @property
    def id(self) -> str:
        self._data._set_default_value_if_none('id')
        return self._data.id

    @id.setter
    def id(self, value: str):
        self._data.id = value

    @property
    def parent_id(self) -> Optional[str]:
        self._data._set_default_value_if_none('parent_id')
        return self._data.parent_id

    @parent_id.setter
    def parent_id(self, value: str):
        self._data.parent_id = value

    @property
    def granularity(self) -> Optional[int]:
        self._data._set_default_value_if_none('granularity')
        return self._data.granularity

    @granularity.setter
    def granularity(self, value: int):
        self._data.granularity = value

    @property
    def adjacency(self) -> Optional[int]:
        self._data._set_default_value_if_none('adjacency')
        return self._data.adjacency

    @adjacency.setter
    def adjacency(self, value: int):
        self._data.adjacency = value

    @property
    def blob(self) -> Optional[bytes]:
        self._data._set_default_value_if_none('blob')
        return self._data.blob

    @blob.setter
    def blob(self, value: bytes):
        self._data.blob = value

    @property
    def tensor(self) -> Optional['ArrayType']:
        self._data._set_default_value_if_none('tensor')
        return self._data.tensor

    @tensor.setter
    def tensor(self, value: 'ArrayType'):
        self._data.tensor = value

    @property
    def mime_type(self) -> Optional[str]:
        self._data._set_default_value_if_none('mime_type')
        return self._data.mime_type

    @mime_type.setter
    def mime_type(self, value: str):
        self._data.mime_type = value

    @property
    def text(self) -> Optional[str]:
        self._data._set_default_value_if_none('text')
        return self._data.text

    @text.setter
    def text(self, value: str):
        self._data.text = value

    @property
    def content(self) -> Optional['DocumentContentType']:
        self._data._set_default_value_if_none('content')
        return self._data.content

    @content.setter
    def content(self, value: 'DocumentContentType'):
        self._data.content = value

    @property
    def weight(self) -> Optional[float]:
        self._data._set_default_value_if_none('weight')
        return self._data.weight

    @weight.setter
    def weight(self, value: float):
        self._data.weight = value

    @property
    def uri(self) -> Optional[str]:
        self._data._set_default_value_if_none('uri')
        return self._data.uri

    @uri.setter
    def uri(self, value: str):
        self._data.uri = value

    @property
    def tags(self) -> Optional[Dict[str, 'StructValueType']]:
        self._data._set_default_value_if_none('tags')
        return self._data.tags

    @tags.setter
    def tags(self, value: Dict[str, 'StructValueType']):
        self._data.tags = value

    @property
    def _metadata(self) -> Optional[Dict[str, 'StructValueType']]:
        self._data._set_default_value_if_none('_metadata')
        return self._data._metadata

    @_metadata.setter
    def _metadata(self, value: Dict[str, 'StructValueType']):
        self._data._metadata = value

    @property
    def offset(self) -> Optional[float]:
        self._data._set_default_value_if_none('offset')
        return self._data.offset

    @offset.setter
    def offset(self, value: float):
        self._data.offset = value

    @property
    def location(self) -> Optional[List[float]]:
        self._data._set_default_value_if_none('location')
        return self._data.location

    @location.setter
    def location(self, value: List[float]):
        self._data.location = value

    @property
    def embedding(self) -> Optional['ArrayType']:
        self._data._set_default_value_if_none('embedding')
        return self._data.embedding

    @embedding.setter
    def embedding(self, value: 'ArrayType'):
        self._data.embedding = value

    @property
    def modality(self) -> Optional[str]:
        self._data._set_default_value_if_none('modality')
        return self._data.modality

    @modality.setter
    def modality(self, value: str):
        self._data.modality = value

    @property
    def evaluations(self) -> Optional[Dict[str, 'NamedScore']]:
        self._data._set_default_value_if_none('evaluations')
        return self._data.evaluations

    @evaluations.setter
    def evaluations(self, value: Dict[str, 'NamedScore']):
        self._data.evaluations = value

    @property
    def scores(self) -> Optional[Dict[str, 'NamedScore']]:
        self._data._set_default_value_if_none('scores')
        return self._data.scores

    @scores.setter
    def scores(self, value: Dict[str, 'NamedScore']):
        self._data.scores = value

    @property
    def chunks(self) -> Optional['ChunkArray']:
        self._data._set_default_value_if_none('chunks')
        return self._data.chunks

    @chunks.setter
    def chunks(self, value: 'DocumentArray'):
        self._data.chunks = value

    @property
    def matches(self) -> Optional['MatchArray']:
        self._data._set_default_value_if_none('matches')
        return self._data.matches

    @matches.setter
    def matches(self, value: 'DocumentArray'):
        self._data.matches = value



from docarray.array.document import DocumentArray
from docarray.array.storage.annlite import StorageMixins, AnnliteConfig

__all__ = ['AnnliteConfig', 'DocumentArrayAnnlite']


class DocumentArrayAnnlite(StorageMixins, DocumentArray):
    """
    DocumentArray that stores Documents in `ANNLite <https://github.com/jina-ai/annlite>`_.

    .. note::
        This DocumentArray requires `annlite`. You can install it via `pip install "docarray[annlite]"`.

    With this implementation, :meth:`match` and :meth:`find` perform fast (approximate) vector search.
    Additionally, search with filters on a :class:`~docarray.document.Document` s :attr:`~docarray.document.Document.tags` attribute is supported.

    Example usage:

    .. code-block:: python

        from docarray import Document, DocumentArray
        import numpy as np

        da = DocumentArray(storage='annlite', config={'data_path': './data', 'n_dim': 10})

        n_dim = 3
        da = DocumentArray(
            storage='annlite',
            config={
                'n_dim': n_dim,
                'columns': [('price', 'float')],
            },
        )

        with da:
            da.extend([Document(id=f'r{i}', tags={'price': i}) for i in range(10)])

        max_price = 3
        n_limit = 4

        filter = {'price': {'$lte': max_price}}
        results = da.find(filter=filter)

    .. seealso::
        For further details, see our :ref:`user guide <annlite>`.
    """

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)



from typing import Union, Iterable

from docarray.array.storage.base.seqlike import BaseSequenceLikeMixin
from docarray.array.memory import DocumentArrayInMemory
from docarray import Document


class SequenceLikeMixin(BaseSequenceLikeMixin):
    """Implement sequence-like methods"""

    def extend(self, values: Iterable['Document']) -> None:
        docs = DocumentArrayInMemory(values)
        if len(docs) == 0:
            return

        for doc in docs:
            doc.embedding = self._map_embedding(doc.embedding)

        self._annlite.index(docs)
        self._offset2ids.extend([doc.id for doc in docs])

        self._update_subindices_append_extend(docs)

    def append(self, value: 'Document'):
        self.extend([value])

    def __del__(self) -> None:
        if not self._persist:
            self._offset2ids.clear()
            self._annlite.clear()

    def __eq__(self, other):
        """In annlite backend, data are considered as identical if configs point to the same database source"""
        return (
            type(self) is type(other)
            and type(self._config) is type(other._config)
            and self._config == other._config
        )

    def __repr__(self):
        return f'<DocumentArray[AnnLite] (length={len(self)}) at {id(self)}>'

    def __contains__(self, x: Union[str, 'Document']):
        if isinstance(x, str):
            return self._annlite.get_doc_by_id(x) is not None
        elif isinstance(x, Document):
            return self._annlite.get_doc_by_id(x.id) is not None
        else:
            return False



import itertools
from typing import (
    TYPE_CHECKING,
    Union,
    Sequence,
    overload,
    Any,
    List,
)

import numpy as np

from docarray import Document
from docarray.helper import typename

if TYPE_CHECKING:
    from docarray.typing import (
        DocumentArrayIndexType,
        DocumentArraySingletonIndexType,
        DocumentArrayMultipleIndexType,
        DocumentArrayMultipleAttributeType,
        DocumentArraySingleAttributeType,
    )


class SetItemMixin:
    """Provides helper function to allow advanced indexing for `__setitem__`"""

    @overload
    def __setitem__(
        self,
        index: 'DocumentArrayMultipleAttributeType',
        value: List[List['Any']],
    ):
        ...

    @overload
    def __setitem__(
        self,
        index: 'DocumentArraySingleAttributeType',
        value: List['Any'],
    ):
        ...

    @overload
    def __setitem__(
        self,
        index: 'DocumentArraySingletonIndexType',
        value: 'Document',
    ):
        ...

    @overload
    def __setitem__(
        self,
        index: 'DocumentArrayMultipleIndexType',
        value: Sequence['Document'],
    ):
        ...

    def __setitem__(
        self,
        index: 'DocumentArrayIndexType',
        value: Union['Document', Sequence['Document']],
    ):

        self._update_subindices_set(index, value)
        # set by offset
        # allows da[1] = Document()
        if isinstance(index, (int, np.generic)) and not isinstance(index, bool):
            self._set_doc_by_offset(int(index), value)
        elif isinstance(index, str):
            # set by traversal paths
            # allows da['@m,c] = [m1, m2, ..., mn, c1, c2, ..., cp]
            if index.startswith('@'):
                self._set_doc_value_pairs_nested(self.traverse_flat(index[1:]), value)

            # set by ID
            # allows da['id_123'] = Document()
            else:
                self._set_doc(index, value)
        # set by slice
        # allows da[1:3] = [d1, d2]
        elif isinstance(index, slice):
            self._set_docs_by_slice(index, value)

        # flatten and set
        # allows da[...] = [d1, d2,..., dn]
        elif index is Ellipsis:
            self._set_doc_value_pairs(self.flatten(), value)

        # index is sequence
        elif isinstance(index, Sequence):
            # allows da[idx1, idx2] = value
            if isinstance(index, tuple) and len(index) == 2:
                self._set_by_pair(index[0], index[1], value)

            # allows da[True, False, True, True]
            elif isinstance(index[0], bool):
                self._set_by_mask(index, value)

            # allows da[id1, id2, id3] = [d1, d2, d3]
            elif isinstance(index[0], (int, str)):
                for si, _val in zip(index, value):
                    self[si] = _val  # leverage existing setter
            else:
                raise IndexError(
                    f'{index} should be either a sequence of bool, int or str'
                )

        # set by ndarray
        elif isinstance(index, np.ndarray):
            index = index.squeeze()
            if index.ndim == 1:
                self[index.tolist()] = value  # leverage existing setter
            else:
                raise IndexError(
                    f'When using np.ndarray as index, its `ndim` must =1. However, receiving ndim={index.ndim}'
                )
        else:
            raise IndexError(f'Unsupported index type {typename(index)}: {index}')

    def _set_by_pair(self, idx1, idx2, value):

        if isinstance(idx1, str) and not idx1.startswith('@'):
            # second is an ID
            # allows da[id1, id2] = [d1, d2]
            if isinstance(idx2, str) and idx2 in self:
                self._set_doc_value_pairs((self[idx1], self[idx2]), value)
            # second is an attribute
            # allows da[id, attr] = attr_value
            elif isinstance(idx2, str) and hasattr(self[idx1], idx2):
                self._set_doc_attr_by_id(idx1, idx2, value)
            # second is a list of attributes:
            # allows da[id, [attr1, attr2, attr3]] = [v1, v2, v3]
            elif (
                isinstance(idx2, Sequence)
                and all(isinstance(attr, str) for attr in idx2)
                and all(hasattr(self[idx1], attr) for attr in idx2)
            ):

                for attr, _v in zip(idx2, value):
                    self._set_doc_attr_by_id(idx1, attr, _v)
            else:
                raise IndexError(f'`{idx2}` is neither a valid id nor attribute name')
        elif isinstance(idx1, int):
            # second is an offset
            # allows da[offset1, offset2] = [d1, d2]
            if isinstance(idx2, int):
                self._set_doc_value_pairs((self[idx1], self[idx2]), value)
            # second is an attribute
            # allows da[offset, attr] = value
            elif isinstance(idx2, str) and hasattr(self[idx1], idx2):
                self._set_doc_attr_by_offset(idx1, idx2, value)
            # second is a list of attributes
            # allows da[offset, [attr1, attr2, attr3]] = [v1, v2, v3]
            elif (
                isinstance(idx2, Sequence)
                and all(isinstance(attr, str) for attr in idx2)
                and all(hasattr(self[idx1], attr) for attr in idx2)
            ):
                for attr, _v in zip(idx2, value):
                    self._set_doc_attr_by_offset(idx1, attr, _v)
            else:
                raise IndexError(f'`{idx2}` must be an attribute or list of attributes')

        # allows da[sequence/slice/ellipsis/traversal_path, attributes] = [v1, v2, ...]
        elif (
            isinstance(idx1, (slice, Sequence))
            or idx1 is Ellipsis
            or (isinstance(idx1, str) and idx1.startswith('@'))
        ):
            self._set_docs_attributes(idx1, idx2, value)
        else:
            raise IndexError(f'Unsupported first index type {typename(idx1)}: {idx1}')

    def _set_by_mask(self, mask: List[bool], value):
        _selected = itertools.compress(self, mask)
        self._set_doc_value_pairs(_selected, value)

    def _set_docs_attributes(self, index, attributes, value):
        if isinstance(attributes, str):
            # a -> [a]
            # [a, a] -> [a, a]
            attributes = (attributes,)
            value = (value,)

        if isinstance(index, str) and index.startswith('@'):
            self._set_docs_attributes_traversal_paths(index, attributes, value)
        elif index is Ellipsis:
            _docs = self[index]
            for _a, _v in zip(attributes, value):
                if _a == 'tensor':
                    _docs.tensors = _v
                elif _a == 'embedding':
                    _docs.embeddings = _v
                else:
                    if not isinstance(_v, (list, tuple)):
                        for _d in _docs:
                            setattr(_d, _a, _v)
                    else:
                        for _d, _vv in zip(_docs, _v):
                            setattr(_d, _a, _vv)
            self._set_doc_value_pairs_nested(_docs, _docs)
        else:
            _docs = self[index]
            if not _docs:
                return

            for _a, _v in zip(attributes, value):
                if _a in ('tensor', 'embedding'):
                    if _a == 'tensor':
                        _docs.tensors = _v
                    elif _a == 'embedding':
                        _docs.embeddings = _v
                    for _d in _docs:
                        self._set_doc(_d.id, _d)
                else:
                    if not isinstance(_v, (list, tuple)):
                        for _d in _docs:
                            self._set_doc_attr_by_id(_d.id, _a, _v)
                    else:
                        for _d, _vv in zip(_docs, _v):
                            self._set_doc_attr_by_id(_d.id, _a, _vv)

    def _set_docs_attributes_traversal_paths(
        self, traversal_paths: str, attributes, value
    ):
        _docs = self[traversal_paths]
        if not _docs:
            return

        for _a, _v in zip(attributes, value):
            if _a == 'tensor':
                _docs.tensors = _v
            elif _a == 'embedding':
                _docs.embeddings = _v
            else:
                if not isinstance(_v, (list, tuple)):
                    for _d in _docs:
                        setattr(_d, _a, _v)
                else:
                    for _d, _vv in zip(_docs, _v):
                        setattr(_d, _a, _vv)
        self._set_doc_value_pairs_nested(_docs, _docs)



from typing import TYPE_CHECKING, Tuple, Sequence, Optional, List, Any

import numpy as np

if TYPE_CHECKING:
    from docarray.typing import ArrayType
    from docarray import Document, DocumentArray


def unravel(docs: Sequence['Document'], field: str) -> Optional['ArrayType']:
    _first = getattr(docs[0], field)
    if _first is None:
        # failed to unravel, return as a list
        r = [getattr(d, field) for d in docs]
        if any(_rr is not None for _rr in r):
            return r
        else:
            return None

    framework, is_sparse = get_array_type(_first)
    cls_type = type(_first)

    all_fields = [getattr(d, field) for d in docs]
    none_idx = [idx for idx, v in enumerate(all_fields) if v is None]
    if none_idx:
        raise ValueError(
            f'Document{none_idx}.{field} is None. Can not stack into `{field}s`.'
        )

    if framework == 'python':
        return cls_type(all_fields)

    elif framework == 'numpy':
        return np.stack(all_fields)

    elif framework == 'tensorflow':
        import tensorflow as tf

        return tf.stack(all_fields)

    elif framework == 'torch':
        import torch

        return torch.stack(all_fields)

    elif framework == 'paddle':
        import paddle

        return paddle.stack(all_fields)

    elif framework == 'scipy':
        import scipy.sparse

        return cls_type(scipy.sparse.vstack(all_fields))


def ravel(value: 'ArrayType', docs: 'DocumentArray', field: str) -> None:
    """Ravel :attr:`value` into ``doc.field`` of each documents

    :param docs: the docs to set
    :param field: the field of the doc to set
    :param value: the value to be set on ``doc.field``
    """
    use_get_row = False
    if hasattr(value, 'getformat'):
        # for scipy only
        sp_format = value.getformat()
        if sp_format in {'bsr', 'coo'}:
            # for BSR and COO, they dont implement [j, ...] in scipy
            # but they offer get_row() API which implicitly translate the
            # sparse row into CSR format, hence needs to convert back
            # not very efficient, but this is the best we can do.
            use_get_row = True

    if use_get_row:
        emb_shape0 = value.shape[0]
        for d, j in zip(docs, range(emb_shape0)):
            row = getattr(value.getrow(j), f'to{sp_format}')()
            docs[d.id, field] = row
    elif isinstance(value, (list, tuple)):
        for d, j in zip(docs, value):
            docs[d.id, field] = j
    else:

        emb_shape0 = value.shape[0]
        for d, j in zip(docs, range(emb_shape0)):
            docs[d.id, field] = value[j, ...]


def get_array_type(
    array: 'ArrayType', raise_error_if_not_array: bool = True
) -> Tuple[str, bool]:
    """Get the type of ndarray without importing the framework

    :param array: any array, scipy, numpy, tf, torch, etc.
    :return: a tuple where the first element represents the framework, the second represents if it is sparse array
    """
    module_tags = array.__class__.__module__.split('.')
    class_name = array.__class__.__name__

    if isinstance(array, (list, tuple)):
        return 'python', False

    if 'numpy' in module_tags:
        return 'numpy', False

    if 'docarray' in module_tags:
        if class_name == 'NdArray':
            return 'docarray', False  # sparse or not is irrelevant

    if 'docarray_pb2' in module_tags:
        if class_name == 'NdArrayProto':
            return 'docarray_proto', False  # sparse or not is irrelevant

    if 'tensorflow' in module_tags:
        if class_name == 'SparseTensor':
            return 'tensorflow', True
        if class_name == 'Tensor' or class_name == 'EagerTensor':
            return 'tensorflow', False

    if 'torch' in module_tags and class_name == 'Tensor':
        return 'torch', array.is_sparse

    if 'paddle' in module_tags and class_name == 'Tensor':
        # Paddle does not support sparse tensor on 11/8/2021
        # https://github.com/PaddlePaddle/Paddle/issues/36697
        return 'paddle', False

    if 'scipy' in module_tags and 'sparse' in module_tags:
        return 'scipy', True

    if raise_error_if_not_array:
        if array is not None:
            raise TypeError(
                f'can not determine the array type: {module_tags}.{class_name}'
            )
        else:
            raise ValueError(
                f'Empty ndarray. Did you forget to set .embedding/.tensor value and now you are operating on it?'
            )
    else:
        return 'python', False


def to_numpy_array(value) -> 'np.ndarray':
    """Return the value always in :class:`numpy.ndarray` regardless the framework type.

    :return: the value in :class:`numpy.ndarray`.
    """
    v = value
    framework, is_sparse = get_array_type(value)
    if is_sparse:
        if hasattr(v, 'todense'):
            v = v.todense()
        elif hasattr(v, 'to_dense'):
            v = v.to_dense()
        elif framework == 'tensorflow':
            import tensorflow as tf

            if isinstance(v, tf.SparseTensor):
                v = tf.sparse.to_dense(v)

    if hasattr(v, 'numpy'):
        v = v.numpy()
    if framework == 'python':
        v = np.array(v)
    return v


def to_list(value) -> List[float]:
    r = to_numpy_array(value)
    if isinstance(r, np.ndarray):
        return r.tolist()
    elif isinstance(r, list):
        return r
    else:
        raise TypeError(f'{r} can not be converted into list')


def get_array_rows(array: 'ArrayType') -> Tuple[int, int]:
    """Get the number of rows of the ndarray without importing all frameworks

    :param array: input array
    :return: (num_rows, ndim)

    Examples

    >>> get_array_rows([1,2,3])
    1, 1
    >>> get_array_rows([[1,2,3], [4,5,6]])
    2, 2
    >>> get_array_rows([[1,2,3], [4,5,6], [7,8,9]])
    3, 2
    >>> get_array_rows(np.array([[1,2,3], [4,5,6], [7,8,9]]))
    3, 2
    """
    array_type, _ = get_array_type(array)

    if array_type == 'python':
        first_element_list_like = isinstance(array[0], (list, tuple))
        num_rows = len(array) if first_element_list_like else 1
        ndim = 2 if first_element_list_like else 1
    elif array_type in ('numpy', 'tensorflow', 'torch', 'paddle', 'scipy'):
        ndim = array.ndim
        if ndim == 1:
            num_rows = 1
        else:
            num_rows = array.shape[0]
    else:
        raise ValueError

    return num_rows, ndim


def check_arraylike_equality(x: 'ArrayType', y: 'ArrayType'):
    """Check if two array type objects are the same with the supported frameworks.

    Examples

    >>> import numpy as np
        x = np.array([[1,2,0,0,3],[1,2,0,0,3]])
        check_arraylike_equality(x,x)
    True

    >>> from scipy import sparse as sp
        x = sp.csr_matrix([[1,2,0,0,3],[1,2,0,0,3]])
        check_arraylike_equality(x,x)
    True

    >>> import torch
        x = torch.tensor([1,2,3])
        check_arraylike_equality(x,x)
    True
    """
    x_type, x_is_sparse = get_array_type(x)
    y_type, y_is_sparse = get_array_type(y)

    same_array = False
    if x_type == y_type and x_is_sparse == y_is_sparse:

        if x_type == 'python':
            same_array = x == y

        if x_type == 'numpy':
            # Numpy does not support sparse tensors
            import numpy as np

            same_array = np.array_equal(x, y)
        elif x_type == 'torch':
            import torch

            if x_is_sparse:
                # torch.equal NotImplementedError for sparse
                same_array = all((x - y).coalesce().values() == 0)
            else:
                same_array = torch.equal(x, y)
        elif x_type == 'scipy':
            # Not implemented in scipy this should work for all types
            # Note: you can't simply look at nonzero values because they can be in
            # different positions.
            if x.shape != y.shape:
                same_array = False
            else:
                same_array = (x != y).nnz == 0
        elif x_type == 'tensorflow':
            if x_is_sparse:
                same_array = x == y
            else:
                # Does not have equal implemented, only elementwise, therefore reduce .all is needed
                same_array = (x == y).numpy().all()
        elif x_type == 'paddle':
            # Paddle does not support sparse tensor on 11/8/2021
            # https://github.com/PaddlePaddle/Paddle/issues/36697
            # Does not have equal implemented, only elementwise, therefore reduce .all is needed
            same_array = (x == y).numpy().all()
        return same_array
    else:
        return same_array


def detach_tensor_if_present(x: Any) -> Any:
    """Check if input is a dense torch array and detaches the tensor from the current graph.
    :param array: input array
    :return: (num_rows, ndim)
    """
    x_type, x_sparse = get_array_type(x, raise_error_if_not_array=False)
    if x_type == 'torch' and not x_sparse:
        import torch

        x = torch.tensor(x.detach().numpy())
    return x



import os
import warnings
from pathlib import Path
from typing import Dict, Type, TYPE_CHECKING, Optional

from docarray.helper import get_request_header, __cache_path__

if TYPE_CHECKING:
    from docarray.typing import T


class PushPullMixin:
    """Transmitting :class:`DocumentArray` via Jina Cloud Service"""

    _max_bytes = 4 * 1024 * 1024 * 1024

    def push(
        self,
        name: str,
        show_progress: bool = False,
        public: bool = True,
    ) -> Dict:
        """Push this DocumentArray object to Jina Cloud which can be later retrieved via :meth:`.push`

        .. note::
            - Push with the same ``name`` will override the existing content.
            - Kinda like a public clipboard where everyone can override anyone's content.
              So to make your content survive longer, you may want to use longer & more complicated name.
            - The lifetime of the content is not promised atm, could be a day, could be a week. Do not use it for
              persistence. Only use this full temporary transmission/storage/clipboard.

        :param name: a name that later can be used for retrieve this :class:`DocumentArray`.
        :param show_progress: if to show a progress bar on pulling
        :param public: by default anyone can pull a DocumentArray if they know its name.
            Setting this to False will allow only the creator to pull it. This feature of course you to login first.
        """
        import requests

        delimiter = os.urandom(32)

        (data, ctype) = requests.packages.urllib3.filepost.encode_multipart_formdata(
            {
                'file': (
                    'DocumentArray',
                    delimiter,
                ),
                'name': name,
                'type': 'documentArray',
                'public': public,
            }
        )

        headers = {'Content-Type': ctype, **get_request_header()}
        import hubble

        auth_token = hubble.get_token()
        if auth_token:
            headers['Authorization'] = f'token {auth_token}'

        if not public and not auth_token:
            warnings.warn(
                'You are not logged in, and `public=False` if only for logged in users. To login, use `jina auth login`.'
            )

        _head, _tail = data.split(delimiter)
        _head += self._stream_header
        from rich import filesize
        from docarray.array.mixins.io.pbar import get_progressbar

        pbar, t = get_progressbar('Pushing', disable=not show_progress, total=len(self))

        def gen():
            total_size = 0

            pbar.start_task(t)

            yield _head

            def _get_chunk(_batch):
                return b''.join(
                    d._to_stream_bytes(protocol='protobuf', compress='gzip')
                    for d in _batch
                ), len(_batch)

            for chunk, num_doc_in_chunk in self.map_batch(_get_chunk, batch_size=32):
                total_size += len(chunk)
                if total_size > self._max_bytes:
                    warnings.warn(
                        f'DocumentArray is too big. The pushed DocumentArray might be chopped off.'
                    )
                    break
                yield chunk
                pbar.update(
                    t,
                    advance=num_doc_in_chunk,
                    total_size=str(filesize.decimal(total_size)),
                )
            yield _tail

        with pbar:
            from hubble import Client
            from hubble.client.endpoints import EndpointsV2

            response = requests.post(
                Client()._base_url + EndpointsV2.upload_artifact,
                data=gen(),
                headers=headers,
            )

        if response.ok:
            return response.json()['data']
        else:
            response.raise_for_status()

    @classmethod
    def pull(
        cls: Type['T'],
        name: str,
        show_progress: bool = False,
        local_cache: bool = True,
        *args,
        **kwargs,
    ) -> 'T':
        """Pulling a :class:`DocumentArray` from Jina Cloud Service to local.

        :param name: the upload name set during :meth:`.push`
        :param show_progress: if to show a progress bar on pulling
        :param local_cache: store the downloaded DocumentArray to local folder
        :return: a :class:`DocumentArray` object
        """

        import requests

        headers = {}

        import hubble

        auth_token = hubble.get_token()

        if auth_token:
            headers['Authorization'] = f'token {auth_token}'

        from hubble import Client
        from hubble.client.endpoints import EndpointsV2

        url = Client()._base_url + EndpointsV2.download_artifact + f'?name={name}'
        response = requests.get(url, headers=headers)

        if response.ok:
            url = response.json()['data']['download']
        else:
            response.raise_for_status()

        with requests.get(
            url,
            stream=True,
            headers=get_request_header(),
        ) as r:
            r.raise_for_status()

            _da_len = int(r.headers['Content-length'])

            from docarray.array.mixins.io.binary import LazyRequestReader

            _source = LazyRequestReader(r)
            if local_cache and os.path.exists(f'{__cache_path__}/{name}'):
                _cache_len = os.path.getsize(f'{__cache_path__}/{name}')
                if _cache_len == _da_len:
                    _source = f'{__cache_path__}/{name}'

            r = cls.load_binary(
                _source,
                protocol='protobuf',
                compress='gzip',
                _show_progress=show_progress,
                *args,
                **kwargs,
            )

            if isinstance(_source, LazyRequestReader) and local_cache:
                Path(__cache_path__).mkdir(parents=True, exist_ok=True)
                with open(f'{__cache_path__}/{name}', 'wb') as fp:
                    fp.write(_source.content)

            return r



from rich.progress import (
    Progress,
    BarColumn,
    SpinnerColumn,
    MofNCompleteColumn,
    TextColumn,
    TimeRemainingColumn,
    Text,
)


class QPSColumn(TextColumn):
    def render(self, task) -> Text:
        if task.speed:
            _text = f'{task.speed:.0f} QPS'
        else:
            _text = 'unknown'
        if self.markup:
            text = Text.from_markup(_text, style=self.style, justify=self.justify)
        else:
            text = Text(_text, style=self.style, justify=self.justify)
        if self.highlighter:
            self.highlighter.highlight(text)
        return text


def get_pbar(disable):
    return Progress(
        SpinnerColumn(),
        TextColumn('[bold]{task.description}'),
        BarColumn(),
        MofNCompleteColumn(),
        '•',
        QPSColumn('{task.speed} QPS', justify='right', style='progress.data.speed'),
        '•',
        TimeRemainingColumn(),
        '•',
        TextColumn(
            '[bold blue]{task.fields[total_size]}',
            justify='right',
            style='progress.filesize',
        ),
        transient=False,
        disable=disable,
    )


def get_progressbar(description, disable, total):
    progress = get_pbar(disable)
    task = progress.add_task(description, total=total, start=False, total_size=0)
    return progress, task



from abc import ABC

from docarray.array.storage.memory.backend import BackendMixin
from docarray.array.storage.memory.find import FindMixin
from docarray.array.storage.memory.getsetdel import GetSetDelMixin
from docarray.array.storage.memory.seqlike import SequenceLikeMixin

__all__ = ['StorageMixins']


class StorageMixins(FindMixin, BackendMixin, GetSetDelMixin, SequenceLikeMixin, ABC):
    ...



from docarray.array.document import DocumentArray
from docarray.array.storage.memory import StorageMixins


class DocumentArrayInMemory(StorageMixins, DocumentArray):
    """
    Default DocumentArray that stores Documents in memory.
    With this implementation, :meth:`match` and :meth:`find` perform exact (exhaustive) vector search.

    Example usage:

    .. code-block:: python

        from docarray import Document, DocumentArray

        da = DocumentArray(
            [Document(text='The cake is a lie'), Document(text='Do a barrel roll!')]
        )
        da.apply(Document.embed_feature_hashing)

        query = Document(text='Can i have some cake?').embed_feature_hashing()
        query.match(da, metric='jaccard', use_scipy=True)

        print(query.matches[:, ('text', 'scores__jaccard__value')])

    .. code-block:: bash

        [['The cake is a lie', 'Do a barrel roll!'], [0.9, 1.0]]

    .. seealso::
        For further details, see our :ref:`user guide <documentarray>`.
    """

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)



from dataclasses import dataclass, fields, field
from typing import Optional, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from docarray.score import NamedScore

default_values = dict(value=0.0, op_name='', description='', ref_id='')


@dataclass(unsafe_hash=True)
class NamedScoreData:
    _reference_ns: 'NamedScore' = field(hash=False, compare=False)
    value: Optional[float] = None
    op_name: Optional[str] = None
    description: Optional[str] = None
    ref_id: Optional[str] = None

    @property
    def _non_empty_fields(self) -> Tuple[str]:
        r = []
        for f in fields(self):
            f_name = f.name
            if not f_name.startswith('_'):
                v = getattr(self, f_name)
                if v is not None:
                    if f_name not in default_values:
                        r.append(f_name)
                    else:
                        dv = default_values[f_name]
                        if v != dv:
                            r.append(f_name)

        return tuple(r)

    def _set_default_value_if_none(self, key):
        if getattr(self, key) is None:
            setattr(self, key, default_values[key])



__version__ = '0.15.0'

import os

from docarray.document import Document
from docarray.array import DocumentArray
from docarray.dataclasses import dataclass, field

if 'DA_RICH_HANDLER' in os.environ:
    from rich.traceback import install

    install()



from pathlib import Path
from typing import Union, Optional, Callable, TYPE_CHECKING, Generator

if TYPE_CHECKING:
    from docarray import DocumentArray
    from docarray.typing import T
    from multiprocessing.pool import ThreadPool, Pool


class DataLoaderMixin:
    @classmethod
    def dataloader(
        cls,
        path: Union[str, Path],
        func: Callable[['DocumentArray'], 'T'],
        batch_size: int,
        protocol: str = 'protobuf',
        compress: Optional[str] = None,
        backend: str = 'thread',
        num_worker: Optional[int] = None,
        pool: Optional[Union['Pool', 'ThreadPool']] = None,
        show_progress: bool = False,
    ) -> Generator['DocumentArray', None, None]:
        """Load array elements, batches and maps them with a function in parallel, finally yield the batch in DocumentArray

        :param path: Path or filename where the data is stored.
        :param func: a function that takes :class:`DocumentArray` as input and outputs anything. You can either modify elements
            in-place (only with `thread` backend) or work later on return elements.
        :param batch_size: Size of each generated batch (except the last one, which might be smaller)
        :param protocol: protocol to use
        :param compress: compress algorithm to use
        :param backend: if to use multi-`process` or multi-`thread` as the parallelization backend. In general, if your
            ``func`` is IO-bound then perhaps `thread` is good enough. If your ``func`` is CPU-bound then you may use `process`.
            In practice, you should try yourselves to figure out the best value. However, if you wish to modify the elements
            in-place, regardless of IO/CPU-bound, you should always use `thread` backend.

            .. warning::
                When using `process` backend, you should not expect ``func`` modify elements in-place. This is because
                the multiprocessing backing pass the variable via pickle and work in another process. The passed object
                and the original object do **not** share the same memory.

        :param num_worker: the number of parallel workers. If not given, then the number of CPUs in the system will be used.
        :param pool: use an existing/external pool. If given, `backend` is ignored and you will be responsible for closing the pool.
        :param show_progress: if set, show a progressbar
        :return:
        """
        from docarray.array.mixins.dataloader.helper import DocumentArrayLoader

        for da in DocumentArrayLoader(
            path, protocol=protocol, compress=compress, show_progress=False
        ).map_batch(
            func,
            batch_size=batch_size,
            backend=backend,
            num_worker=num_worker,
            pool=pool,
            show_progress=show_progress,
        ):
            yield da



import uuid
from dataclasses import dataclass, field, asdict
from typing import (
    Iterable,
    Dict,
    Optional,
    TYPE_CHECKING,
    Union,
    Tuple,
    List,
)

import numpy as np
import weaviate

from docarray import Document
from docarray.helper import dataclass_from_dict, filter_dict, _safe_cast_int
from docarray.array.storage.base.backend import BaseBackendMixin, TypeMap
from docarray.array.storage.registry import _REGISTRY

if TYPE_CHECKING:
    from docarray.typing import ArrayType, DocumentArraySourceType


@dataclass
class WeaviateConfig:
    """This class stores the config variables to initialize
    connection to the Weaviate server"""

    host: Optional[str] = field(default='localhost')
    port: Optional[int] = field(default=8080)
    protocol: Optional[str] = field(default='http')
    name: Optional[str] = None
    serialize_config: Dict = field(default_factory=dict)
    n_dim: Optional[int] = None  # deprecated, not used anymore since weaviate 1.10
    # vectorIndexConfig parameters
    ef: Optional[int] = None
    ef_construction: Optional[int] = None
    timeout_config: Optional[Tuple[int, int]] = None
    max_connections: Optional[int] = None
    dynamic_ef_min: Optional[int] = None
    dynamic_ef_max: Optional[int] = None
    dynamic_ef_factor: Optional[int] = None
    vector_cache_max_objects: Optional[int] = None
    flat_search_cutoff: Optional[int] = None
    cleanup_interval_seconds: Optional[int] = None
    skip: Optional[bool] = None
    columns: Optional[List[Tuple[str, str]]] = None
    distance: Optional[str] = None


_banned_classname_chars = [
    '[',
    ' ',
    '"',
    '*',
    '\\',
    '<',
    '|',
    ',',
    '>',
    '/',
    '?',
    ']',
    '@',
    '.',
]


def _sanitize_class_name(name):
    new_name = name
    for char in _banned_classname_chars:
        new_name = new_name.replace(char, '')
    return new_name


class BackendMixin(BaseBackendMixin):
    """Provide necessary functions to enable this storage backend."""

    TYPE_MAP = {
        'str': TypeMap(type='string', converter=str),
        'float': TypeMap(type='number', converter=float),
        'int': TypeMap(type='int', converter=_safe_cast_int),
    }

    def _init_storage(
        self,
        _docs: Optional['DocumentArraySourceType'] = None,
        config: Optional[Union[WeaviateConfig, Dict]] = None,
        **kwargs,
    ):
        """Initialize weaviate storage.

        :param docs: the list of documents to initialize to
        :param config: the config object used to ininitialize connection to weaviate server
        :param kwargs: extra keyword arguments
        :raises ValueError: only one of name or docs can be used for initialization,
            raise an error if both are provided
        """

        if not config:
            config = WeaviateConfig()
        elif isinstance(config, dict):
            config = dataclass_from_dict(WeaviateConfig, config)

        self._serialize_config = config.serialize_config

        if config.name and config.name != config.name.capitalize():
            raise ValueError(
                'Weaviate class name has to be capitalized. '
                'Please capitalize when declaring the name field in config.'
            )

        self._persist = bool(config.name)

        self._client = weaviate.Client(
            f'{config.protocol}://{config.host}:{config.port}',
            timeout_config=config.timeout_config,
        )
        self._config = config

        self._config.columns = self._normalize_columns(self._config.columns)

        self._schemas = self._load_or_create_weaviate_schema()

        _REGISTRY[self.__class__.__name__][self._class_name].append(self)

        super()._init_storage(_docs, **kwargs)

        # To align with Sqlite behavior; if `docs` is not `None` and table name
        # is provided, :class:`DocumentArraySqlite` will clear the existing
        # table and load the given `docs`
        if _docs is None:
            return
        elif isinstance(_docs, Iterable):
            self.clear()
            self.extend(_docs)
        else:
            self.clear()
            if isinstance(_docs, Document):
                self.append(_docs)

    def _ensure_unique_config(
        self,
        config_root: dict,
        config_subindex: dict,
        config_joined: dict,
        subindex_name: str,
    ) -> dict:
        if 'name' not in config_subindex:
            unique_name = _sanitize_class_name(
                config_joined['name'] + 'subindex' + subindex_name
            )
            config_joined['name'] = unique_name
        return config_joined

    def _get_weaviate_class_name(self) -> str:
        """Generate the class/schema name using the ``uuid1`` module with some
        formatting to tailor to weaviate class name convention

        :return: string representing the name of  weaviate class/schema name of
            this :class:`DocumentArrayWeaviate` object
        """
        return f'Class{uuid.uuid4().hex}'

    def _get_schema_by_name(self, cls_name: str) -> Dict:
        """Return the schema dictionary object with the class name
        Content of the all dictionaries by this method are the same except the name
        of the weaviate's ``class``

        :param cls_name: the name of the schema/class in weaviate
        :return: the schema dictionary
        """
        # TODO: ideally we should only use one schema. this will allow us to deal with
        # consistency better
        hnsw_config = {
            'ef': self._config.ef,
            'efConstruction': self._config.ef_construction,
            'maxConnections': self._config.max_connections,
            'dynamicEfMin': self._config.dynamic_ef_min,
            'dynamicEfMax': self._config.dynamic_ef_max,
            'dynamicEfFactor': self._config.dynamic_ef_factor,
            'vectorCacheMaxObjects': self._config.vector_cache_max_objects,
            'flatSearchCutoff': self._config.flat_search_cutoff,
            'cleanupIntervalSeconds': self._config.cleanup_interval_seconds,
            'skip': self._config.skip,
            'distance': self._config.distance,
        }

        base_classes = {
            'classes': [
                {
                    'class': cls_name,
                    "vectorizer": "none",
                    'vectorIndexConfig': {'skip': False, **filter_dict(hnsw_config)},
                    'properties': [
                        {
                            'dataType': ['blob'],
                            'name': '_serialized',
                            'indexInverted': False,
                        },
                    ],
                },
                {
                    'class': cls_name + 'Meta',
                    "vectorizer": "none",
                    'vectorIndexConfig': {'skip': True},
                    'properties': [
                        {
                            'dataType': ['string[]'],
                            'name': '_offset2ids',
                            'indexInverted': False,
                        },
                    ],
                },
            ]
        }
        for col, coltype in self._config.columns:
            new_property = {
                'dataType': [self._map_type(coltype)],
                'name': col,
                'indexInverted': True,
            }
            base_classes['classes'][0]['properties'].append(new_property)

        return base_classes

    def _load_or_create_weaviate_schema(self):
        """Create a new weaviate schema for this :class:`DocumentArrayWeaviate` object
        if not present in weaviate or if ``self._config.name`` is None. If ``self._config.name``
        is provided and not None and schema with the specified name exists in weaviate,
        then load the object with the given ``self._config.name``

        :return: the schemas of this :class`DocumentArrayWeaviate` object and its meta
        """
        if not self._config.name:
            name_candidate = self._get_weaviate_class_name()
            doc_schemas = self._get_schema_by_name(name_candidate)
            while self._client.schema.contains(doc_schemas):
                name_candidate = self._get_weaviate_class_name()
                doc_schemas = self._get_schema_by_name(name_candidate)
            self._client.schema.create(doc_schemas)
            self._config.name = name_candidate
            return doc_schemas

        doc_schemas = self._get_schema_by_name(self._config.name)
        if self._client.schema.contains(doc_schemas):
            return doc_schemas

        self._client.schema.create(doc_schemas)
        return doc_schemas

    def _update_offset2ids_meta(self):
        """Update the offset2ids in weaviate the the current local version"""
        if self._offset2ids_wid is not None and self._client.data_object.exists(
            self._offset2ids_wid
        ):
            self._client.data_object.update(
                data_object={'_offset2ids': self._offset2ids.ids},
                class_name=self._meta_name,
                uuid=self._offset2ids_wid,
            )
        else:
            self._offset2ids_wid = str(uuid.uuid1())
            self._client.data_object.create(
                data_object={'_offset2ids': self._offset2ids.ids},
                class_name=self._meta_name,
                uuid=self._offset2ids_wid,
            )

    def _get_offset2ids_meta(self) -> Tuple[List, str]:
        """Return the offset2ids stored in weaviate along with the name of the schema/class
        in weaviate that stores meta information of this object

        :return: a tuple with first element as a list of offset2ids and second element
                 being name of weaviate class/schema of the meta object

        :raises ValueError: error is raised if meta class name is not defined
        """
        if not self._meta_name:
            raise ValueError('meta object is not defined')

        resp = (
            self._client.query.get(self._meta_name, ['_offset2ids', '_additional {id}'])
            .do()
            .get('data', {})
            .get('Get', {})
            .get(self._meta_name, [])
        )

        if not resp:
            return [], None
        elif len(resp) == 1:
            return resp[0]['_offset2ids'], resp[0]['_additional']['id']
        else:
            raise ValueError('received multiple meta copies which is invalid')

    @property
    def name(self):
        """An alias to _class_name that returns the id/name of the class
        in the weaviate of this :class:`DocumentArrayWeaviate`

        :return: name of weaviate class/schema of this :class:`DocumentArrayWeaviate`
        """
        return self._class_name

    @property
    def _class_name(self):
        """Return the name of the class in weaviate of this :class:`DocumentArrayWeaviate

        :return: name of weaviate class/schema of this :class:`DocumentArrayWeaviate`
        """
        if not self._schemas:
            return None
        return self._schemas['classes'][0]['class']

    @property
    def _meta_name(self):
        """Return the name of the class in weaviate that stores the meta information of
        this :class:`DocumentArrayWeaviate`

        :return: name of weaviate class/schema of class that stores the meta information
        """
        # TODO: remove this after we combine the meta info to the DocumentArray class
        if not self._schemas:
            return None
        return self._schemas['classes'][1]['class']

    @property
    def _class_schema(self) -> Optional[Dict]:
        """Return the schema dictionary of this :class:`DocumentArrayWeaviate`'s weaviate schema

        :return: the dictionary representing this weaviate schema
        """
        if not self._schemas:
            return None
        return self._schemas['classes'][0]

    @property
    def _meta_schema(self):
        """Return the schema dictionary of this weaviate schema that stores this object's meta

        :return: the dictionary representing a meta object's weaviate schema
        """
        if not self._schemas and len(self._schemas) < 2:
            return None
        return self._schemas['classes'][1]

    def _doc2weaviate_create_payload(self, value: 'Document'):
        """Return the payload to store :class:`Document` into weaviate

        :param value: document to create a payload for
        :return: the payload dictionary
        """
        columns_dict = {key: val for [key, val] in self._config.columns}
        extra_columns = {
            col: self._map_column(value.tags.get(col), columns_dict[col])
            for col, _ in self._config.columns
        }

        return dict(
            data_object={
                '_serialized': value.to_base64(**self._serialize_config),
                **extra_columns,
            },
            class_name=self._class_name,
            uuid=self._map_id(value.id),
            vector=self._map_embedding(value.embedding),
        )

    def _map_id(self, doc_id: str):
        """the function maps doc id to weaviate id

        :param doc_id: id of the document
        :return: weaviate object id
        """
        # appending class name to doc id to handle the case:
        # daw1 = DocumentArrayWeaviate([Document(id=str(i), text='hi') for i in range(3)])
        # daw2 = DocumentArrayWeaviate([Document(id=str(i), text='bye') for i in range(3)])
        # daw2[0, 'text'] == 'hi' # this will be False if we don't append class name
        return str(uuid.uuid5(uuid.NAMESPACE_URL, doc_id + self._class_name))

    def _map_embedding(self, embedding: 'ArrayType'):
        if embedding is not None:
            from docarray.math.ndarray import to_numpy_array

            embedding = to_numpy_array(embedding)

            if embedding.ndim > 1:
                embedding = np.asarray(embedding).squeeze()

            # Weaviate expects vector to have dim 2 at least
            # or get weaviate.exceptions.UnexpectedStatusCodeException:  models.C11yVector
            # hence we cast it to list of a single element
            if len(embedding) == 1:
                embedding = [embedding[0]]
        else:
            embedding = None
        return embedding

    def __getstate__(self):
        d = dict(self.__dict__)
        del d['_client']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self._client = weaviate.Client(
            f'{state["_config"].protocol}://{state["_config"].host}:{state["_config"].port}'
        )



from abc import ABC, abstractmethod
from collections import namedtuple
from dataclasses import is_dataclass, asdict
from typing import Dict, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from docarray.typing import DocumentArraySourceType, ArrayType

TypeMap = namedtuple('TypeMap', ['type', 'converter'])


class BaseBackendMixin(ABC):
    TYPE_MAP: Dict[str, TypeMap]

    def _init_storage(
        self,
        _docs: Optional['DocumentArraySourceType'] = None,
        copy: bool = False,
        *args,
        **kwargs,
    ):
        self._load_offset2ids()

    def _init_subindices(self, *args, **kwargs):
        self._subindices = {}
        subindex_configs = kwargs.get('subindex_configs', None)
        if subindex_configs:
            config = asdict(self._config) if getattr(self, '_config', None) else dict()

            for name, config_subindex in subindex_configs.items():
                config_subindex = (
                    dict() if config_subindex is None else config_subindex
                )  # allow None as input
                config_joined = {**config, **config_subindex}
                config_joined = self._ensure_unique_config(
                    config, config_subindex, config_joined, name
                )
                self._subindices[name] = self.__class__(config=config_joined)
                self._subindices[name].extend(self.traverse_flat(name[1:]))

    @abstractmethod
    def _ensure_unique_config(
        self,
        config_root: dict,
        config_subindex: dict,
        config_joined: dict,
        subindex_name: str,
    ) -> dict:
        """
        Ensures that the subindex configuration is unique, despite it inheriting unpopulated fields from the root config.

        :param config_root: The configuration of the root index.
        :param config_subindex: The configuration that was explicitly provided by the user for the subindex.
        :param config_joined: The configuration that combines root and subindex configs. This is the configuration that will be used for subindex construction.
        :param subindex_name: Name (access path) of the subindex
        :return: config_joined that is unique compared to config_root
        """
        ...

    def _get_storage_infos(self) -> Optional[Dict]:
        if hasattr(self, '_config') and is_dataclass(self._config):
            return {k: str(v) for k, v in asdict(self._config).items()}

    def _map_id(self, _id: str) -> str:
        return _id

    def _map_column(self, value, col_type) -> str:
        return self.TYPE_MAP[col_type].converter(value)

    def _map_embedding(self, embedding: 'ArrayType') -> 'ArrayType':
        from docarray.math.ndarray import to_numpy_array

        return to_numpy_array(embedding)

    def _map_type(self, col_type: str) -> str:
        return self.TYPE_MAP[col_type].type

    def _normalize_columns(self, columns):
        if columns is None:
            return []
        return columns



from typing import TYPE_CHECKING, Type, List

if TYPE_CHECKING:
    from docarray.typing import T
    from docarray.document.strawberry_type import StrawberryDocument


class StrawberryMixin:
    def to_strawberry_type(self) -> List['StrawberryDocument']:
        """Convert a DocumentArray object into a Pydantic model."""
        return [d.to_strawberry_type() for d in self]

    @classmethod
    def from_strawberry_type(cls: Type['T'], model: List['StrawberryDocument']) -> 'T':
        """Convert a list of Strawberry into DocumentArray

        :param model: the list of strawberry type objects that represents a DocumentArray
        :return: a DocumentArray
        """
        from docarray import Document

        return cls(Document.from_strawberry_type(m) for m in model)



from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from docarray.typing import ArrayType


def cosine(x_mat: 'np.ndarray', y_mat: 'np.ndarray', eps: float = 1e-7) -> 'np.ndarray':
    """Cosine distance between each row in x_mat and each row in y_mat.

    :param x_mat: np.ndarray with ndim=2
    :param y_mat: np.ndarray with ndim=2
    :param eps: a small jitter to avoid divde by zero
    :return: np.ndarray  with ndim=2
    """
    return 1 - np.clip(
        (np.dot(x_mat, y_mat.T) + eps)
        / (
            np.outer(np.linalg.norm(x_mat, axis=1), np.linalg.norm(y_mat, axis=1)) + eps
        ),
        -1,
        1,
    )


def sqeuclidean(x_mat: 'np.ndarray', y_mat: 'np.ndarray') -> 'np.ndarray':
    """Squared Euclidean distance between each row in x_mat and each row in y_mat.
    :param x_mat: np.ndarray with ndim=2
    :param y_mat: np.ndarray with ndim=2
    :return: np.ndarray with ndim=2
    """
    return (
        np.sum(y_mat**2, axis=1)
        + np.sum(x_mat**2, axis=1)[:, np.newaxis]
        - 2 * np.dot(x_mat, y_mat.T)
    )


def sparse_cosine(x_mat: 'ArrayType', y_mat: 'ArrayType') -> 'np.ndarray':
    """Cosine distance between each row in x_mat and each row in y_mat.
    :param x_mat:  scipy.sparse like array with ndim=2
    :param y_mat:  scipy.sparse like array with ndim=2
    :return: np.ndarray  with ndim=2
    """
    from scipy.sparse.linalg import norm

    # we need the np.asarray otherwise we get a np.matrix object that iterates differently
    return 1 - np.clip(
        np.asarray(
            x_mat.dot(y_mat.T) / (np.outer(norm(x_mat, axis=1), norm(y_mat, axis=1)))
        ),
        -1,
        1,
    )


def sparse_sqeuclidean(x_mat: 'ArrayType', y_mat: 'ArrayType') -> 'np.ndarray':
    """Cosine distance between each row in x_mat and each row in y_mat.

    :param x_mat:  scipy.sparse like array with ndim=2
    :param y_mat:  scipy.sparse like array with ndim=2
    :return: np.ndarray  with ndim=2
    """
    # we need the np.asarray otherwise we get a np.matrix object that iterates differently
    return np.asarray(
        y_mat.power(2).sum(axis=1).flatten()
        + x_mat.power(2).sum(axis=1)
        - 2 * x_mat.dot(y_mat.T)
    )


def sparse_euclidean(x_mat: 'ArrayType', y_mat: 'ArrayType') -> 'np.ndarray':
    """Sparse euclidean distance between each row in x_mat and each row in y_mat.

    :param x_mat:  scipy.sparse like array with ndim=2
    :param y_mat:  scipy.sparse like array with ndim=2
    :return: np.ndarray  with ndim=2
    """
    return np.sqrt(sparse_sqeuclidean(x_mat, y_mat))


def euclidean(x_mat: 'ArrayType', y_mat: 'ArrayType') -> 'np.ndarray':
    """Euclidean distance between each row in x_mat and each row in y_mat.

    :param x_mat:  scipy.sparse like array with ndim=2
    :param y_mat:  scipy.sparse like array with ndim=2
    :return: np.ndarray  with ndim=2
    """
    return np.sqrt(sqeuclidean(x_mat, y_mat))



from typing import Optional, List, Dict, Any, TYPE_CHECKING, Union

from pydantic import BaseModel, validator

from docarray.math.ndarray import to_list

if TYPE_CHECKING:
    from docarray.typing import ArrayType

# this order must be preserved: https://pydantic-docs.helpmanual.io/usage/types/#unions
_ProtoValueType = Optional[Union[bool, float, str, list, dict]]
_StructValueType = Union[
    _ProtoValueType, List[_ProtoValueType], Dict[str, _ProtoValueType]
]
_MetadataType = Dict[str, _StructValueType]


def _convert_ndarray_to_list(v: 'ArrayType'):
    if v is not None:
        return to_list(v)


class _NamedScore(BaseModel):
    value: Optional[float] = None
    op_name: Optional[str] = None
    description: Optional[str] = None
    ref_id: Optional[str] = None


class _MetadataModel(BaseModel):
    metadata: _MetadataType


class PydanticDocument(BaseModel):
    id: Optional[str]
    parent_id: Optional[str]
    granularity: Optional[int]
    adjacency: Optional[int]
    blob: Optional[str]
    tensor: Optional[Any]
    mime_type: Optional[str]
    text: Optional[str]
    weight: Optional[float]
    uri: Optional[str]
    tags: Optional[Dict[str, '_StructValueType']]
    _metadata: Optional[Dict[str, '_StructValueType']]
    offset: Optional[float]
    location: Optional[List[float]]
    embedding: Optional[Any]
    modality: Optional[str]
    evaluations: Optional[Dict[str, '_NamedScore']]
    scores: Optional[Dict[str, '_NamedScore']]
    chunks: Optional[List['PydanticDocument']]
    matches: Optional[List['PydanticDocument']]

    _tensor2list = validator('tensor', allow_reuse=True)(_convert_ndarray_to_list)
    _embedding2list = validator('embedding', allow_reuse=True)(_convert_ndarray_to_list)

    class Config:
        smart_union = True

    def __init__(self, **data):
        super().__init__(**data)
        # underscore attributes need to be set and validated manually
        _metadata = data.get('_metadata', None)
        if _metadata is not None:
            _md_model = _MetadataModel(metadata=_metadata)  # validate _metadata
            object.__setattr__(self, '_metadata', _md_model.metadata)


PydanticDocument.update_forward_refs()

PydanticDocumentArray = List[PydanticDocument]



from typing import MutableSequence, TYPE_CHECKING, Union, Iterable

from docarray import Document

if TYPE_CHECKING:
    from docarray.typing import T


class BaseDocumentArray(MutableSequence[Document]):
    def __init__(self, *args, storage: str = 'memory', **kwargs):
        super().__init__()
        self._init_storage(*args, **kwargs)
        self._init_subindices(*args, **kwargs)

    def __add__(self: 'T', other: Union['Document', Iterable['Document']]) -> 'T':
        v = type(self)(self)
        v.extend(other)
        return v



from typing import Optional, overload, TYPE_CHECKING, Dict, Union

from docarray.array.base import BaseDocumentArray
from docarray.array.mixins import AllMixins

if TYPE_CHECKING:
    from docarray.typing import DocumentArraySourceType
    from docarray.array.memory import DocumentArrayInMemory
    from docarray.array.sqlite import DocumentArraySqlite
    from docarray.array.annlite import DocumentArrayAnnlite
    from docarray.array.weaviate import DocumentArrayWeaviate
    from docarray.array.elastic import DocumentArrayElastic
    from docarray.array.storage.sqlite import SqliteConfig
    from docarray.array.storage.annlite import AnnliteConfig
    from docarray.array.storage.weaviate import WeaviateConfig
    from docarray.array.storage.elastic import ElasticConfig


class DocumentArray(AllMixins, BaseDocumentArray):
    """
    DocumentArray is a list-like container of :class:`~docarray.Document` objects.

    A DocumentArray can be used to store, embed, and retrieve :class:`~docarray.Document` objects.

    .. code-block:: python

        from docarray import Document, DocumentArray

        da = DocumentArray(
            [Document(text='The cake is a lie'), Document(text='Do a barrel roll!')]
        )
        da.apply(Document.embed_feature_hashing)

        query = Document(text='Can i have some cake?').embed_feature_hashing()
        query.match(da, metric='jaccard', use_scipy=True)

        print(query.matches[:, ('text', 'scores__jaccard__value')])

    .. code-block:: bash

        [['The cake is a lie', 'Do a barrel roll!'], [0.9, 1.0]]

    A DocumentArray can also :ref:`embed its contents using a neural network <embed-via-model>`,
    process them using an :ref:`external Flow or Executor <da-post>`, and persist Documents in a :ref:`Document Store <doc-store>` for
    fast vector search:

    .. code-block:: python

        from docarray import Document, DocumentArray
        import numpy as np

        n_dim = 3
        metric = 'Euclidean'

        # initialize a DocumentArray with ANNLiter Document Store
        da = DocumentArray(
            storage='annlite',
            config={'n_dim': n_dim, 'columns': [('price', 'float')], 'metric': metric},
        )
        # add Documents to the DocumentArray
        with da:
            da.extend(
                [
                    Document(id=f'r{i}', embedding=i * np.ones(n_dim), tags={'price': i})
                    for i in range(10)
                ]
            )
        # perform vector search
        np_query = np.ones(n_dim) * 8
        results = da.find(np_query)

    .. seealso::
        For further details, see our :ref:`user guide <documentarray>`.
    """

    @overload
    def __new__(
        cls,
        _docs: Optional['DocumentArraySourceType'] = None,
        copy: bool = False,
        subindex_configs: Optional[Dict[str, 'None']] = None,
    ) -> 'DocumentArrayInMemory':
        """Create an in-memory DocumentArray object."""
        ...

    @overload
    def __new__(
        cls,
        _docs: Optional['DocumentArraySourceType'] = None,
        storage: str = 'sqlite',
        config: Optional[Union['SqliteConfig', Dict]] = None,
        subindex_configs: Optional[Dict[str, Dict]] = None,
    ) -> 'DocumentArraySqlite':
        """Create a SQLite-powered DocumentArray object."""
        ...

    @overload
    def __new__(
        cls,
        _docs: Optional['DocumentArraySourceType'] = None,
        storage: str = 'weaviate',
        config: Optional[Union['WeaviateConfig', Dict]] = None,
        subindex_configs: Optional[Dict[str, Dict]] = None,
    ) -> 'DocumentArrayWeaviate':
        """Create a Weaviate-powered DocumentArray object."""
        ...

    @overload
    def __new__(
        cls,
        _docs: Optional['DocumentArraySourceType'] = None,
        storage: str = 'annlite',
        config: Optional[Union['AnnliteConfig', Dict]] = None,
        subindex_configs: Optional[Dict[str, Dict]] = None,
    ) -> 'DocumentArrayAnnlite':
        """Create a AnnLite-powered DocumentArray object."""
        ...

    @overload
    def __new__(
        cls,
        _docs: Optional['DocumentArraySourceType'] = None,
        storage: str = 'elasticsearch',
        config: Optional[Union['ElasticConfig', Dict]] = None,
        subindex_configs: Optional[Dict[str, Dict]] = None,
    ) -> 'DocumentArrayElastic':
        """Create a Elastic-powered DocumentArray object."""
        ...

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        """
        Ensures that offset2ids are stored in the db after
        operations in the DocumentArray are performed.
        """
        self._save_offset2ids()

    def __new__(cls, *args, storage: str = 'memory', **kwargs):
        if cls is DocumentArray:
            if storage == 'memory':
                from docarray.array.memory import DocumentArrayInMemory

                instance = super().__new__(DocumentArrayInMemory)
            elif storage == 'sqlite':
                from docarray.array.sqlite import DocumentArraySqlite

                instance = super().__new__(DocumentArraySqlite)
            elif storage == 'annlite':
                from docarray.array.annlite import DocumentArrayAnnlite

                instance = super().__new__(DocumentArrayAnnlite)
            elif storage == 'weaviate':
                from docarray.array.weaviate import DocumentArrayWeaviate

                instance = super().__new__(DocumentArrayWeaviate)
            elif storage == 'qdrant':
                from docarray.array.qdrant import DocumentArrayQdrant

                instance = super().__new__(DocumentArrayQdrant)
            elif storage == 'elasticsearch':
                from docarray.array.elastic import DocumentArrayElastic

                instance = super().__new__(DocumentArrayElastic)

            else:
                raise ValueError(f'storage=`{storage}` is not supported.')
        else:
            instance = super().__new__(cls)
        return instance



from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:
    from pandas import DataFrame
    from docarray.typing import T


class DataframeIOMixin:
    """Save/load from :class:`pandas.dataframe`

    .. note::
        These functions require you to install `pandas`

    """

    def to_dataframe(self, **kwargs) -> 'DataFrame':
        """Export itself to a :class:`pandas.DataFrame` object.

        :param kwargs: the extra kwargs will be passed to :meth:`pandas.DataFrame.from_dict`.
        :return: a :class:`pandas.DataFrame` object
        """
        from pandas import DataFrame

        return DataFrame.from_dict(self.to_list(), **kwargs)

    @classmethod
    def from_dataframe(cls: Type['T'], df: 'DataFrame', *args, **kwargs) -> 'T':
        """Import a :class:`DocumentArray` from a :class:`pandas.DataFrame` object.

        :param df: a :class:`pandas.DataFrame` object.
        :return: a :class:`DocumentArray` object
        """
        da = cls(**kwargs)
        from docarray import Document

        for m in df.to_dict(orient='records'):
            # drop nan
            da.append(
                Document(
                    {k: v for k, v in m.items() if (not isinstance(v, float) or v == v)}
                )
            )
        return da



from typing import overload, TYPE_CHECKING, Union, Callable, Optional, Tuple

if TYPE_CHECKING:
    from docarray import DocumentArray
    from docarray.typing import AnyDNN, T, ArrayType

    import numpy as np


class SingletonSugarMixin:
    """Provide sugary syntax for :class:`Document` by inheriting methods from :class:`DocumentArray`"""

    @overload
    def match(
        self: 'T',
        darray: 'DocumentArray',
        metric: Union[
            str, Callable[['ArrayType', 'ArrayType'], 'np.ndarray']
        ] = 'cosine',
        limit: Optional[Union[int, float]] = 20,
        normalization: Optional[Tuple[float, float]] = None,
        metric_name: Optional[str] = None,
        batch_size: Optional[int] = None,
        exclude_self: bool = False,
        only_id: bool = False,
        use_scipy: bool = False,
        num_worker: Optional[int] = 1,
    ) -> 'T':
        """Matching the current Document against a set of Documents.

        The result will be stored in :attr:`.matches`.

        .. note::
            When you want to match a set Documents (let's call it set `A`) against another set of Documents (set `B`),
            where you want to find for each element in `A` what are its nearest neighbours in `B`.
            Then you need :meth:`DocumentArray.match`

        :param darray: the other DocumentArray to match against
        :param metric: the distance metric
        :param limit: the maximum number of matches, when not given defaults to 20.
        :param normalization: a tuple [a, b] to be used with min-max normalization,
                                the min distance will be rescaled to `a`, the max distance will be rescaled to `b`
                                all values will be rescaled into range `[a, b]`.
        :param metric_name: if provided, then match result will be marked with this string.
        :param batch_size: if provided, then ``darray`` is loaded in batches, where each of them is at most ``batch_size``
            elements. When `darray` is big, this can significantly speedup the computation.
        :param exclude_self: if set, Documents in ``darray`` with same ``id`` as the left-hand values will not be
                        considered as matches.
        :param only_id: if set, then returning matches will only contain ``id``
        :param use_scipy: if set, use ``scipy`` as the computation backend. Note, ``scipy`` does not support distance
            on sparse matrix.
        :param num_worker: the number of parallel workers. If not given, then the number of CPUs in the system will be used.

                .. note::
                    This argument is only effective when ``batch_size`` is set.
        """
        ...

    def match(self: 'T', *args, **kwargs) -> 'T':
        from docarray import DocumentArray

        _tmp = DocumentArray(self)
        _tmp.match(*args, **kwargs)
        return self

    @overload
    def embed(
        self: 'T',
        embed_model: 'AnyDNN',
        device: str = 'cpu',
        batch_size: int = 256,
    ) -> 'T':
        """Fill the embedding of Documents inplace by using `embed_model`

        :param embed_model: the embedding model written in Keras/Pytorch/Paddle
        :param device: the computational device for `embed_model`, can be either
            `cpu` or `cuda`.
        :param batch_size: number of Documents in a batch for embedding
        """

    def embed(self: 'T', *args, **kwargs) -> 'T':
        from docarray import DocumentArray

        _tmp = DocumentArray(self)
        _tmp.embed(*args, **kwargs)
        return self

    def post(self: 'T', *args, **kwargs) -> 'T':
        from docarray import DocumentArray

        _tmp = DocumentArray(self)
        return _tmp.post(*args, **kwargs)[0]



from docarray.array.document import DocumentArray

from docarray.array.storage.sqlite import StorageMixins, SqliteConfig

__all__ = ['SqliteConfig', 'DocumentArraySqlite']


class DocumentArraySqlite(StorageMixins, DocumentArray):
    """
    DocumentArray that stores Documents in a `SQLite database <https://www.sqlite.org/index.html>`_.
    This stores Documents on disk instead of keeping them in memory, and offers the simplest way of persisting data with DocArray.

    With this implementation, :meth:`match` and :meth:`find` perform exact (exhaustive) vector search.

    Example usage:

    .. code-block:: python

        from docarray import DocumentArray

        # with default config
        da = DocumentArray(storage='sqlite')

        # with customized config
        da1 = DocumentArray(storage='sqlite', config={'connection': 'example.db'})

        # connect to a previously created database
        da = DocumentArray(
            storage='sqlite', config={'connection': 'example.db', 'table_name': 'mine'}
        )


    .. seealso::
        For further details, see our :ref:`user guide <sqlite>`.
    """

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)



from typing import (
    TYPE_CHECKING,
    Iterable,
)

from docarray.array.memory import DocumentArrayInMemory

if TYPE_CHECKING:
    from docarray.document import Document


class MatchArray(DocumentArrayInMemory):
    """
    :class:`MatchArray` inherits from :class:`DocumentArray`.
    It's a subset of Documents that represents the matches

    :param docs: Set of matches of the `reference_doc`
    :param reference_doc: Reference :class:`Document` for the sub-documents
    """

    def __init__(self, docs, reference_doc: 'Document'):
        self._ref_doc = reference_doc
        super().__init__(docs)
        if isinstance(docs, Iterable) and self._ref_doc is not None:
            for d in docs:
                d.adjacency = self._ref_doc.adjacency + 1

    def append(self, document: 'Document'):
        """Add a matched document to the current Document.

        :param document: Sub-document to be added
        """
        document.adjacency = self._ref_doc.adjacency + 1
        super().append(document)

    @property
    def reference_doc(self) -> 'Document':
        """Get the document that this :class:`MatchArray` referring to.
        :return: the document the match refers to
        """
        return self._ref_doc

    @property
    def granularity(self) -> int:
        """Get granularity of all document in this array.
        :return: the granularity of the documents of which these are match
        """
        return self._ref_doc.granularity

    @property
    def adjacency(self) -> int:
        """Get the adjacency of all document in this array.
        :return: the adjacency of the array of matches
        """
        return self._ref_doc.adjacency + 1



from typing import TYPE_CHECKING

import paddle

if TYPE_CHECKING:
    from paddle import tensor
    import numpy


def cosine(
    x_mat: 'tensor', y_mat: 'tensor', eps: float = 1e-7, device: str = 'cpu'
) -> 'numpy.ndarray':
    """Cosine distance between each row in x_mat and each row in y_mat.

    :param x_mat: np.ndarray with ndim=2
    :param y_mat: np.ndarray with ndim=2
    :param eps: a small jitter to avoid divde by zero
    :param device: the computational device for `embed_model`, can be either `cpu` or `cuda`.
    :return: np.ndarray  with ndim=2
    """
    paddle.set_device(device)

    a_n, b_n = x_mat.norm(axis=1)[:, None], y_mat.norm(axis=1)[:, None]
    a_norm = x_mat / paddle.clip(a_n, min=eps)
    b_norm = y_mat / paddle.clip(b_n, min=eps)
    sim_mt = 1 - paddle.mm(a_norm, b_norm.transpose(perm=[1, 0]))
    return sim_mt.numpy()


def sqeuclidean(
    x_mat: 'tensor', y_mat: 'tensor', device: str = 'cpu'
) -> 'numpy.ndarray':
    """Squared euclidean distance between each row in x_mat and each row in y_mat.

    :param x_mat:  paddle array with ndim=2
    :param y_mat:  paddle array with ndim=2
    :param device: the computational device for `embed_model`, can be either `cpu` or `cuda`.
    :return: np.ndarray  with ndim=2
    """
    paddle.set_device(device)

    return (
        paddle.sum(y_mat**2, axis=1)
        + paddle.sum(x_mat**2, axis=1)[:, None]
        - 2 * paddle.mm(x_mat, y_mat.transpose(perm=[1, 0]))
    ).numpy()


def euclidean(x_mat: 'tensor', y_mat: 'tensor', device: str = 'cpu') -> 'numpy.ndarray':
    """Euclidean distance between each row in x_mat and each row in y_mat.

    :param x_mat:  paddle array with ndim=2
    :param y_mat:  paddle array with ndim=2
    :param device: the computational device for `embed_model`, can be either `cpu` or `cuda`.
    :return: np.ndarray  with ndim=2
    """
    paddle.set_device(device)

    return paddle.sqrt(
        paddle.sum(y_mat**2, axis=1)
        + paddle.sum(x_mat**2, axis=1)[:, None]
        - 2 * paddle.mm(x_mat, y_mat.transpose(perm=[1, 0]))
    ).numpy()



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from docarray import Document


def image_getter(doc: 'Document'):
    if doc._metadata['image_type'] == 'uri':
        return doc.uri
    elif doc._metadata['image_type'] == 'PIL':
        from PIL import Image

        return Image.fromarray(doc.tensor)
    elif doc._metadata['image_type'] == 'ndarray':
        return doc.tensor


def text_getter(doc: 'Document'):
    return doc.text


def audio_getter(doc: 'Document'):
    return doc.uri or doc.tensor


def video_getter(doc: 'Document'):
    return doc.uri or doc.tensor


def mesh_getter(doc: 'Document'):
    return doc.uri or doc.tensor


def tabular_getter(doc: 'Document'):
    return doc.uri


def blob_getter(doc: 'Document'):
    return doc.uri or doc.blob


def json_getter(doc: 'Document'):
    return doc.tags



from collections import namedtuple
from typing import TYPE_CHECKING, Dict, NamedTuple, Optional
from urllib.parse import urlparse

if TYPE_CHECKING:
    from docarray import DocumentArray


_ParsedHost = namedtuple('ParsedHost', 'on host port version scheme')


def _parse_host(host: str) -> NamedTuple:
    """Parse a host string into namedtuple object.

    A parsed host's components are `on`, `host`, `port`, `version`, `scheme`.
    :param host: a host string. Can be one of the following:
        - `grpc://192.168.0.123:8080/endpoint`
        - `ws://192.168.0.123:8080/endpoint`
        - `http://192.168.0.123:8080/endpoint`
        - `jinahub://Hello/endpoint`
        - `jinahub+docker://Hello/endpoint`
        - `jinahub+docker://Hello/v0.0.1/endpoint`
        - `jinahub+docker://Hello/latest/endpoint`
        - `jinahub+sandbox://Hello/endpoint`
    """
    r = urlparse(host)
    on = r.path or '/'
    host = (
        r._replace(netloc=r.netloc.replace(f':{r.port}', ''))._replace(path='').geturl()
    )
    port = r.port or None
    version = None
    scheme = r.scheme
    splited_path = list(filter(None, r.path.split('/')))
    if len(splited_path) == 2:
        # path includes version and endpoint
        version = splited_path[0]
        host = host + '/' + version
        on = '/' + splited_path[1]

    return _ParsedHost(on=on, host=host, port=port, version=version, scheme=scheme)


class PostMixin:
    """Helper functions for posting DocumentArray to Jina Flow."""

    def post(
        self,
        host: str,
        show_progress: bool = False,
        batch_size: Optional[int] = None,
        parameters: Optional[Dict] = None,
        **kwargs,
    ) -> 'DocumentArray':
        """Posting itself to a remote Flow/Sandbox and get the modified DocumentArray back

        :param host: a host string. Can be one of the following:
            - `grpc://192.168.0.123:8080/endpoint`
            - `ws://192.168.0.123:8080/endpoint`
            - `http://192.168.0.123:8080/endpoint`
            - `jinahub://Hello/endpoint`
            - `jinahub+docker://Hello/endpoint`
            - `jinahub+docker://Hello/v0.0.1/endpoint`
            - `jinahub+docker://Hello/latest/endpoint`
            - `jinahub+sandbox://Hello/endpoint`

        :param show_progress: if to show a progressbar
        :param batch_size: number of Document on each request
        :param parameters: parameters to send in the request
        :return: the new DocumentArray returned from remote
        """

        if not self:
            return

        parsed_host = _parse_host(host)

        batch_size = batch_size or len(self)

        scheme = parsed_host.scheme
        host = parsed_host.host

        if scheme in ('grpcs', 'https', 'wss'):
            scheme = scheme[:-1]

        if scheme == 'ws':
            scheme = 'websocket'  # temp fix for the core

        if scheme.startswith('jinahub'):
            from jina import Flow

            f = Flow(quiet=True, prefetch=1).add(uses=host, **kwargs)
            with f:
                return f.post(
                    parsed_host.on,
                    inputs=self,
                    show_progress=show_progress,
                    request_size=batch_size,
                    parameters=parameters,
                    **kwargs,
                )
        elif scheme in ('grpc', 'http', 'ws', 'websocket'):
            from jina import Client

            if parsed_host.port:
                host += f':{parsed_host.port}'

            c = Client(host=host)
            return c.post(
                parsed_host.on,
                inputs=self,
                show_progress=show_progress,
                request_size=batch_size,
                parameters=parameters,
                **kwargs,
            )
        else:
            raise ValueError(f'unsupported scheme: {scheme}')



from docarray.array.document import DocumentArray
from docarray.array.storage.weaviate import StorageMixins, WeaviateConfig

__all__ = ['DocumentArrayWeaviate', 'WeaviateConfig']


class DocumentArrayWeaviate(StorageMixins, DocumentArray):
    """
    DocumentArray that stores Documents in a `Weaviate <https://weaviate.io/>`_ vector search engine.

    .. note::
        This DocumentArray requires `weaviate-client`. You can install it via `pip install "docarray[weaviate]"`.

        To use Weaviate as storage backend, a Weaviate service needs to be running on your machine.

    With this implementation, :meth:`match` and :meth:`find` perform fast (approximate) vector search.
    Additionally, search with filters is supported.

    Example usage:

    .. code-block:: python

        from docarray import DocumentArray

        # connect to running Weaviate service with default configuration (address: http://localhost:8080)
        da = DocumentArray(storage='weaviate')

        # connect to a previously persisted DocumentArrayWeaviate by specifying name, host, and port
        da = DocumentArray(
            storage='weaviate', config={'name': 'Persisted', 'host': 'localhost', 'port': 1234}
        )


    .. seealso::
        For further details, see our :ref:`user guide <weaviate>`.
    """

    def __new__(cls, *args, **kwargs):
        """``__new__`` method for :class:`DocumentArrayWeaviate`

        :param *args: list of args to instantiate the object
        :param **kwargs: dict of args to instantiate the object
        :return: the instantiated :class:`DocumentArrayWeaviate` object
        """
        return super().__new__(cls)



from typing import overload, Dict, Optional, List, TYPE_CHECKING, Sequence, Any

from docarray.document.data import DocumentData
from docarray.document.mixins import AllMixins
from docarray.base import BaseDCType
from docarray.math.ndarray import detach_tensor_if_present

if TYPE_CHECKING:
    from docarray.typing import ArrayType, StructValueType, DocumentContentType


class Document(AllMixins, BaseDCType):
    """Document is the basic data type in DocArray.
    A Document is a container for any kind of data, be it text, image, audio, video, or 3D meshes.

    You can initialize a Document object with given attributes:

    .. code-block:: python

        from docarray import Document
        import numpy

        d1 = Document(text='hello')
        d3 = Document(tensor=numpy.array([1, 2, 3]))
        d4 = Document(
            uri='https://jina.ai',
            mime_type='text/plain',
            granularity=1,
            adjacency=3,
            tags={'foo': 'bar'},
        )

    Documents support a :ref:`nested structure <recursive-nested-document>`, which can also be specified during construction:

    .. code-block:: python

        d = Document(
            id='d0',
            chunks=[Document(id='d1', chunks=Document(id='d2'))],
            matches=[Document(id='d3')],
        )

    A Document can embed its contents using the :meth:`embed` method and a provided embedding model:

    .. code-block:: python

        import torchvision

        q = (
            Document(uri='/Users/usr/path/to/image.jpg')
            .load_uri_to_image_tensor()
            .set_image_tensor_normalization()
            .set_image_tensor_channel_axis(-1, 0)
        )
        model = torchvision.models.resnet50(pretrained=True)
        q.embed(model)

    Multiple Documents can be organized into a :class:`~docarray.array.document.DocumentArray`.

    .. seealso::
        For further details, see our :ref:`user guide <document>`.
    """

    _data_class = DocumentData
    _unresolved_fields_dest = 'tags'
    _post_init_fields = (
        'text',
        'blob',
        'tensor',
        'content',
        'uri',
        'mime_type',
        'chunks',
        'matches',
    )

    @overload
    def __init__(self):
        """Create an empty Document."""
        ...

    @overload
    def __init__(self, _obj: Optional['Document'] = None, copy: bool = False):
        ...

    @overload
    def __init__(self, _obj: Optional[Any] = None):
        """Create a Document from a `docarray.dataclass` instance"""
        ...

    @overload
    def __init__(
        self,
        _obj: Optional[Dict],
        copy: bool = False,
        field_resolver: Optional[Dict[str, str]] = None,
        unknown_fields_handler: str = 'catch',
    ):
        ...

    @overload
    def __init__(self, blob: Optional[bytes] = None, **kwargs):
        """Create a Document with binary content."""
        ...

    @overload
    def __init__(self, tensor: Optional['ArrayType'] = None, **kwargs):
        """Create a Document with NdArray-like content."""
        ...

    @overload
    def __init__(self, text: Optional[str] = None, **kwargs):
        """Create a Document with string content."""
        ...

    @overload
    def __init__(self, uri: Optional[str] = None, **kwargs):
        """Create a Document with content from a URI."""
        ...

    @overload
    def __init__(
        self,
        parent_id: Optional[str] = None,
        granularity: Optional[int] = None,
        adjacency: Optional[int] = None,
        blob: Optional[bytes] = None,
        tensor: Optional['ArrayType'] = None,
        mime_type: Optional[str] = None,
        text: Optional[str] = None,
        content: Optional['DocumentContentType'] = None,
        weight: Optional[float] = None,
        uri: Optional[str] = None,
        tags: Optional[Dict[str, 'StructValueType']] = None,
        offset: Optional[float] = None,
        location: Optional[List[float]] = None,
        embedding: Optional['ArrayType'] = None,
        modality: Optional[str] = None,
        evaluations: Optional[Dict[str, Dict[str, 'StructValueType']]] = None,
        scores: Optional[Dict[str, Dict[str, 'StructValueType']]] = None,
        chunks: Optional[Sequence['Document']] = None,
        matches: Optional[Sequence['Document']] = None,
    ):
        ...

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getstate__(self):
        state = self.__dict__.copy()

        for attribute in ['embedding', 'tensor']:
            if hasattr(self, attribute):
                setattr(
                    state['_data'],
                    attribute,
                    detach_tensor_if_present(getattr(state['_data'], attribute)),
                )

        return state



from typing import Type, TYPE_CHECKING

from docarray import Document

if TYPE_CHECKING:
    from docarray.typing import T


class EmptyMixin:
    """Helper functions for building arrays with empty Document."""

    @classmethod
    def empty(cls: Type['T'], size: int = 0, *args, **kwargs) -> 'T':
        """Create a :class:`DocumentArray`  object with :attr:`size` empty
        :class:`Document` objects.

        :param size: the number of empty Documents in this container
        :return: a :class:`DocumentArray` object
        """
        return cls((Document() for _ in range(size)), *args, **kwargs)



from docarray.document.mixins.attribute import GetAttributesMixin
from docarray.document.mixins.audio import AudioDataMixin
from docarray.document.mixins.blob import BlobDataMixin
from docarray.document.mixins.content import ContentPropertyMixin
from docarray.document.mixins.convert import ConvertMixin
from docarray.document.mixins.dump import UriFileMixin
from docarray.document.mixins.featurehash import FeatureHashMixin
from docarray.document.mixins.image import ImageDataMixin
from docarray.document.mixins.mesh import MeshDataMixin
from docarray.document.mixins.multimodal import MultiModalMixin
from docarray.document.mixins.plot import PlotMixin
from docarray.document.mixins.porting import PortingMixin
from docarray.document.mixins.property import PropertyMixin
from docarray.document.mixins.protobuf import ProtobufMixin
from docarray.document.mixins.pydantic import PydanticMixin
from docarray.document.mixins.strawberry import StrawberryMixin
from docarray.document.mixins.sugar import SingletonSugarMixin
from docarray.document.mixins.text import TextDataMixin
from docarray.document.mixins.video import VideoDataMixin


class AllMixins(
    ProtobufMixin,
    PydanticMixin,
    StrawberryMixin,
    PropertyMixin,
    ContentPropertyMixin,
    ConvertMixin,
    AudioDataMixin,
    ImageDataMixin,
    TextDataMixin,
    MeshDataMixin,
    VideoDataMixin,
    BlobDataMixin,
    PlotMixin,
    UriFileMixin,
    SingletonSugarMixin,
    PortingMixin,
    FeatureHashMixin,
    GetAttributesMixin,
    MultiModalMixin,
):
    """All plugins that can be used in :class:`Document`."""

    ...



# auto-generated from /Users/hanxiao/Documents/docarray/scripts/gen_ns_property_mixin.py
from typing import Optional


class PropertyMixin:
    @property
    def value(self) -> Optional[float]:
        self._data._set_default_value_if_none('value')
        return self._data.value

    @value.setter
    def value(self, value: float):
        self._data.value = value

    @property
    def op_name(self) -> Optional[str]:
        self._data._set_default_value_if_none('op_name')
        return self._data.op_name

    @op_name.setter
    def op_name(self, value: str):
        self._data.op_name = value

    @property
    def description(self) -> Optional[str]:
        self._data._set_default_value_if_none('description')
        return self._data.description

    @description.setter
    def description(self, value: str):
        self._data.description = value

    @property
    def ref_id(self) -> Optional[str]:
        self._data._set_default_value_if_none('ref_id')
        return self._data.ref_id

    @ref_id.setter
    def ref_id(self, value: str):
        self._data.ref_id = value



from typing import Union, Iterable, Dict, List
import warnings

from docarray.array.storage.base.seqlike import BaseSequenceLikeMixin
from docarray import Document


class SequenceLikeMixin(BaseSequenceLikeMixin):
    """Implement sequence-like methods for DocumentArray with Elastic as storage"""

    def __eq__(self, other):
        """Compare this object to the other, returns True if and only if other
        as the same type as self and other has the same meta information

        :param other: the other object to check for equality
        :return: ``True`` if other is equal to self
        """
        # two DAW are considered as the same if they have the same client meta data
        return (
            type(self) is type(other)
            and self._client.get_meta() == other._client.get_meta()
            and self._config == other._config
        )

    def __len__(self):
        """Return the length of :class:`DocumentArray` that uses Elastic as storage

        :return: the length of this :class:`DocumentArrayElastic` object
        """
        try:
            return self._client.count(index=self._config.index_name)["count"]
        except:
            return 0

    def __contains__(self, x: Union[str, 'Document']):
        """Check if ``x`` is contained in this :class:`DocumentArray` with Elastic storage

        :param x: the id of the document to check or the document object itself
        :return: True if ``x`` is contained in self
        """
        if isinstance(x, str):
            return self._doc_id_exists(x)
        elif isinstance(x, Document):
            return self._doc_id_exists(x.id)
        else:
            return False

    def __del__(self):
        """Delete this :class:`DocumentArrayElastic` object"""
        self._save_offset2ids()

        # if not self._persist:
        #    self._offset2ids.clear()

    def __repr__(self):
        """Return the string representation of :class:`DocumentArrayElastic` object
        :return: string representation of this object
        """
        return f'<{self.__class__.__name__} (length={len(self)}) at {id(self)}>'

    @staticmethod
    def _parse_index_ids_from_bulk_info(
        accumulated_info: List[Dict],
    ) -> Dict[str, List[int]]:
        """Parse ids from bulk info of failed send request to ES operation

        :param accumulated_info: accumulated info of failed operation
        :return: dict containing failed index ids of each operation type
        """

        parsed_ids = {}

        for info in accumulated_info:
            for _op_type in info.keys():
                if '_id' in info[_op_type]:
                    if _op_type not in parsed_ids:
                        parsed_ids[_op_type] = []

                    parsed_ids[_op_type].append(info[_op_type]['_id'])

        return parsed_ids

    def _upload_batch(self, docs: Iterable['Document'], **kwargs) -> List[int]:
        requests = [self._document_to_elastic(doc) for doc in docs]
        accumulated_info = self._send_requests(requests, **kwargs)
        self._refresh(self._config.index_name)

        successful_ids = self._parse_index_ids_from_bulk_info(accumulated_info)
        if 'index' not in successful_ids:
            return []

        return successful_ids['index']

    def _extend(self, docs: Iterable['Document'], **kwargs):
        docs = list(docs)
        successful_indexed_ids = self._upload_batch(docs, **kwargs)
        self._offset2ids.extend(
            [_id for _id in successful_indexed_ids if _id not in self._offset2ids.ids]
        )

        if len(successful_indexed_ids) != len(docs):
            doc_ids = [doc.id for doc in docs]
            failed_index_ids = set(doc_ids) - set(successful_indexed_ids)

            err_msg = f'fail to add Documents with ids: {failed_index_ids}'
            warnings.warn(err_msg)
            raise IndexError(err_msg)



from docarray.score.data import NamedScoreData
from docarray.score.mixins import AllMixins
from docarray.base import BaseDCType


class NamedScore(AllMixins, BaseDCType):
    _data_class = NamedScoreData
    _post_init_fields = ()



from abc import ABC

from docarray.array.storage.sqlite.backend import BackendMixin, SqliteConfig
from docarray.array.storage.sqlite.getsetdel import GetSetDelMixin
from docarray.array.storage.sqlite.seqlike import SequenceLikeMixin
from docarray.array.storage.memory.find import (
    FindMixin,
)  # temporary delegate to in-memory find API

__all__ = ['StorageMixins', 'SqliteConfig']


class StorageMixins(FindMixin, BackendMixin, GetSetDelMixin, SequenceLikeMixin, ABC):
    ...



from abc import ABC

from docarray.array.storage.elastic.backend import BackendMixin, ElasticConfig
from docarray.array.storage.elastic.find import FindMixin
from docarray.array.storage.elastic.getsetdel import GetSetDelMixin
from docarray.array.storage.elastic.seqlike import SequenceLikeMixin

__all__ = ['StorageMixins', 'ElasticConfig']


class StorageMixins(FindMixin, BackendMixin, GetSetDelMixin, SequenceLikeMixin, ABC):
    ...



from typing import (
    Type,
    TYPE_CHECKING,
    Optional,
    overload,
    Union,
    List,
    TextIO,
    Dict,
    Iterable,
)

if TYPE_CHECKING:
    from docarray.typing import T
    import numpy as np
    import csv


class FromGeneratorMixin:
    """Provide helper functions filling a :class:`DocumentArray`-like object with a generator."""

    @classmethod
    def _from_generator(cls: Type['T'], meth: str, *args, **kwargs) -> 'T':
        from docarray.document import generators

        from_fn = getattr(generators, meth)
        da_like = cls(**kwargs)
        da_like.extend(from_fn(*args, **kwargs))
        return da_like

    @classmethod
    @overload
    def from_ndarray(
        cls: Type['T'],
        array: 'np.ndarray',
        axis: int = 0,
        size: Optional[int] = None,
        shuffle: bool = False,
        *args,
        **kwargs,
    ) -> 'T':
        """Build from a numpy array.

        :param array: the numpy ndarray data source
        :param axis: iterate over that axis
        :param size: the maximum number of the sub arrays
        :param shuffle: shuffle the numpy data source beforehand
        """
        ...

    @classmethod
    def from_ndarray(cls: Type['T'], *args, **kwargs) -> 'T':
        """
        # noqa: DAR101
        # noqa: DAR102
        # noqa: DAR201
        """
        return cls._from_generator('from_ndarray', *args, **kwargs)

    @classmethod
    @overload
    def from_files(
        cls: Type['T'],
        patterns: Union[str, List[str]],
        recursive: bool = True,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
        read_mode: Optional[str] = None,
        to_dataturi: bool = False,
        exclude_regex: Optional[str] = None,
        *args,
        **kwargs,
    ) -> 'T':
        """Build from a list of file path or the content of the files.

        :param patterns: The pattern may contain simple shell-style wildcards, e.g. '\*.py', '[\*.zip, \*.gz]'
        :param recursive: If recursive is true, the pattern '**' will match any files
            and zero or more directories and subdirectories
        :param size: the maximum number of the files
        :param sampling_rate: the sampling rate between [0, 1]
        :param read_mode: specifies the mode in which the file is opened.
            'r' for reading in text mode, 'rb' for reading in binary mode.
            If `read_mode` is None, will iterate over filenames.
        :param to_dataturi: if set, then the Document.uri will be filled with DataURI instead of the plan URI
        :param exclude_regex: if set, then filenames that match to this pattern are not included.

        """
        ...

    @classmethod
    def from_files(cls: Type['T'], *args, **kwargs) -> 'T':
        """
        # noqa: DAR101
        # noqa: DAR102
        # noqa: DAR201
        """
        return cls._from_generator('from_files', *args, **kwargs)

    @classmethod
    @overload
    def from_csv(
        cls: Type['T'],
        file: Union[str, TextIO],
        field_resolver: Optional[Dict[str, str]] = None,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
        dialect: Union[str, 'csv.Dialect'] = 'excel',
    ) -> 'T':
        """Build from CSV.

        :param file: file paths or file handler
        :param field_resolver: a map from field names defined in JSON, dict to the field
                names defined in Document.
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param dialect: define a set of parameters specific to a particular CSV dialect. could be a string that represents
            predefined dialects in your system, or could be a :class:`csv.Dialect` class that groups specific formatting
            parameters together. If you don't know the dialect and the default one does not work for you,
            you can try set it to ``auto``.
        """
        ...

    @classmethod
    def from_csv(cls: Type['T'], *args, **kwargs) -> 'T':
        """
        # noqa: DAR101
        # noqa: DAR102
        # noqa: DAR201
        """
        return cls._from_generator('from_csv', *args, **kwargs)

    @classmethod
    @overload
    def from_huggingface_datasets(
        cls: Type['T'],
        dataset_path: str,
        field_resolver: Optional[Dict[str, str]] = None,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
        filter_fields: bool = False,
        **datasets_kwargs,
    ) -> 'T':
        """Build from Hugging Face Datasets. Yields documents.

        This function helps to load datasets from Hugging Face Datasets Hub
        (https://huggingface.co/datasets) in Jina. Additional parameters can be
        passed to the ``datasets`` library using keyword arguments. The ``load_dataset``
        method from ``datasets`` library is used to load the datasets.

        :param dataset_path: a valid dataset path for Hugging Face Datasets library.
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
                names defined in Protobuf. This is only used when the given ``document`` is
                a JSON string or a Python dict.
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param filter_fields: specifies whether to filter the dataset with the fields
                given in ```field_resolver`` argument.
        :param **datasets_kwargs: additional arguments for ``load_dataset`` method
                from Datasets library. More details at
                https://huggingface.co/docs/datasets/package_reference/loading_methods.html#datasets.load_dataset
        """
        ...

    @classmethod
    def from_huggingface_datasets(cls: Type['T'], *args, **kwargs) -> 'T':
        """
        # noqa: DAR101
        # noqa: DAR102
        # noqa: DAR201
        """
        return cls._from_generator('from_huggingface_datasets', *args, **kwargs)

    @classmethod
    @overload
    def from_ndjson(
        cls: Type['T'],
        fp: Iterable[str],
        field_resolver: Optional[Dict[str, str]] = None,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
    ) -> 'T':
        """Build from line separated JSON. Yields documents.

        :param fp: file paths
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
                names defined in Protobuf. This is only used when the given ``document`` is
                a JSON string or a Python dict.
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        """
        ...

    @classmethod
    def from_ndjson(cls: Type['T'], *args, **kwargs) -> 'T':
        """
        # noqa: DAR101
        # noqa: DAR102
        # noqa: DAR201
        """
        return cls._from_generator('from_ndjson', *args, **kwargs)

    @classmethod
    @overload
    def from_lines(
        cls: 'T',
        lines: Optional[Iterable[str]] = None,
        filepath: Optional[str] = None,
        read_mode: str = 'r',
        line_format: str = 'json',
        field_resolver: Optional[Dict[str, str]] = None,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
    ) -> 'T':
        """Build from lines, json and csv. Yields documents or strings.

        :param lines: a list of strings, each is considered as a document
        :param filepath: a text file that each line contains a document
        :param read_mode: specifies the mode in which the file
                    is opened. 'r' for reading in text mode, 'rb' for reading in binary
        :param line_format: the format of each line ``json`` or ``csv``
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
                names defined in Protobuf. This is only used when the given ``document`` is
                a JSON string or a Python dict.
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        """
        ...

    @classmethod
    def from_lines(cls: Type['T'], *args, **kwargs) -> 'T':
        """
        # noqa: DAR101
        # noqa: DAR102
        # noqa: DAR201
        """
        return cls._from_generator('from_lines', *args, **kwargs)



from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from docarray.typing import T, Document


def _reduce_doc_props(doc1: 'Document', doc2: 'Document'):
    doc1_fields = set(doc1.non_empty_fields)
    doc2_fields = set(doc2.non_empty_fields)

    # update only fields that are set in doc2 and not set in doc1
    fields = doc2_fields - doc1_fields

    fields = fields - {'matches', 'chunks', 'id', 'parent_id'}
    for field in fields:
        setattr(doc1, field, getattr(doc2, field))


class ReduceMixin:
    """
    A mixin that provides reducing logic for :class:`DocumentArray`
    Reducing 2 or more DocumentArrays consists in merging all Documents into the same DocumentArray.
    If a Document belongs to 2 or more DocumentArrays, it is added once and data attributes are merged with priority to
    the Document belonging to the left-most DocumentArray. Matches and chunks are also reduced in the same way.
    Reduction is applied to all levels of DocumentArrays, that is, from root Documents to all their chunk and match
    children.
    """

    def reduce(self: 'T', other: 'T') -> 'T':
        """
        Reduces other and the current DocumentArray into one DocumentArray in-place. Changes are applied to the current
        DocumentArray.
        Reducing 2 DocumentArrays consists in adding Documents in the second DocumentArray to the first DocumentArray
        if they do not exist. If a Document exists in both DocumentArrays, the data properties are merged with priority
        to the first Document (that is, to the current DocumentArray's Document). The matches and chunks are also
        reduced in the same way.
        :param other: DocumentArray
        :return: DocumentArray
        """
        for doc in other:
            if doc.id in self:
                self._reduce_doc(self[doc.id], doc)
            else:
                self.append(doc)

        return self

    @staticmethod
    def _reduce_doc(doc1: 'Document', doc2: 'Document'):
        """
        Reduces doc1 and doc2 into one Document in-place. Changes are applied to doc1.
        Reducing 2 Documents consists in setting data properties of the second Document to the first Document if they
        are empty (that is, priority to the left-most Document) and reducing the matches and the chunks of both
        documents.
        Non-data properties are ignored.
        Reduction of matches and chunks relies on :class:`DocumentArray`.:method:`reduce`.
        :param doc1: first Document
        :param doc2: second Document
        """
        _reduce_doc_props(doc1, doc2)
        if len(doc2.matches) > 0:
            doc1.matches.reduce(doc2.matches)

        if len(doc2.chunks) > 0:
            doc1.chunks.reduce(doc2.chunks)

    def reduce_all(self: 'T', others: List['T']) -> 'T':
        """
        Reduces a list of DocumentArrays and this DocumentArray into one DocumentArray. Changes are applied to this
        DocumentArray in-place.

        Reduction consists in reducing this DocumentArray with every DocumentArray in `others` sequentially using
        :class:`DocumentArray`.:method:`reduce`.
        The resulting DocumentArray contains Documents of all DocumentArrays.
        If a Document exists in many DocumentArrays, data properties are merged with priority to the left-most
        DocumentArrays (that is, if a data attribute is set in a Document belonging to many DocumentArrays, the
        attribute value of the left-most DocumentArray is kept).
        Matches and chunks of a Document belonging to many DocumentArrays are also reduced in the same way.
        Other non-data properties are ignored.

        .. note::
            - Matches are not kept in a sorted order when they are reduced. You might want to re-sort them in a later
                step.
            - The final result depends on the order of DocumentArrays when applying reduction.

        :param others: List of DocumentArrays to be reduced
        :return: the resulting DocumentArray
        """
        for da in others:
            self.reduce(da)
        return self



from abc import ABC

from docarray.array.storage.weaviate.backend import BackendMixin, WeaviateConfig
from docarray.array.storage.weaviate.find import FindMixin
from docarray.array.storage.weaviate.getsetdel import GetSetDelMixin
from docarray.array.storage.weaviate.seqlike import SequenceLikeMixin

__all__ = ['StorageMixins', 'WeaviateConfig']


class StorageMixins(FindMixin, BackendMixin, GetSetDelMixin, SequenceLikeMixin, ABC):
    ...



import copy
from typing import Optional

import numpy as np

from docarray.helper import deprecate_by


class PlotMixin:
    """Provide helper functions for :class:`Document` to plot and visualize itself."""

    def _ipython_display_(self):
        """Displays the object in IPython as a side effect"""
        self.summary()

    def __rich_console__(self, console, options):

        yield f":page_facing_up: [b]Document[/b]: [cyan]{self.id}[cyan]"
        from rich.table import Table
        from rich import box

        my_table = Table(
            'Attribute', 'Value', width=80, box=box.ROUNDED, highlight=True
        )
        for f in self.non_empty_fields:
            if f.startswith('_'):
                continue
            elif f in ('text', 'blob', 'uri') and len(getattr(self, f)) > 100:
                v = getattr(self, f)
                my_table.add_row(f, str(v)[:100] + f'... [dim](length: {len(v)})')
            elif f in ('embedding', 'tensor'):
                from docarray.math.ndarray import to_numpy_array

                v = to_numpy_array(getattr(self, f))
                if v.squeeze().ndim == 1 and len(v) < 1000:
                    from docarray.document.mixins.rich_embedding import (
                        ColorBoxEmbedding,
                    )

                    v = ColorBoxEmbedding(v.squeeze())
                else:
                    v = f'{type(getattr(self, f))} in shape {v.shape}, dtype: {v.dtype}'
                my_table.add_row(f, v)
            elif f not in ('id', 'chunks', 'matches'):
                my_table.add_row(f, str(getattr(self, f)))
        if my_table.rows:
            yield my_table

    def summary(self) -> None:
        """Print non-empty fields and nested structure of this Document object."""
        from rich import print

        print(self._plot_recursion())

    def _plot_recursion(self, tree=None):
        if tree is None:
            from rich.tree import Tree

            tree = Tree(self)
        else:
            tree = tree.add(self)
        for a in ('matches', 'chunks'):
            if getattr(self, a):
                if a == 'chunks':
                    _icon = ':diamond_with_a_dot:'
                else:
                    _icon = ':large_orange_diamond:'
                _match_tree = tree.add(f'{_icon} [b]{a.capitalize()}[/b]')
                for d in getattr(self, a):
                    d._plot_recursion(_match_tree)
        return tree

    def display(self):
        """Plot image data from :attr:`.tensor` or :attr:`.uri`."""
        from IPython.display import Image, display

        if self.uri:
            if self.mime_type.startswith('audio') or self.uri.startswith('data:audio/'):
                uri = _convert_display_uri(self.uri, self.mime_type)
                _html5_audio_player(uri)
            elif self.mime_type.startswith('video') or self.uri.startswith(
                'data:video/'
            ):
                uri = _convert_display_uri(self.uri, self.mime_type)
                _html5_video_player(uri)
            elif self.uri.startswith('data:image/'):
                _html5_image(self.uri)
            else:
                display(Image(self.uri))
        elif self.tensor is not None:
            try:
                import PIL.Image

                p = PIL.Image.fromarray(self.tensor)
                if p.mode != 'RGB':
                    raise
                display(p)
            except:
                import matplotlib.pyplot as plt

                plt.matshow(self.tensor)
        else:
            self.summary()

    def plot_matches_sprites(
        self,
        top_k: int = 10,
        channel_axis: int = -1,
        inv_normalize: bool = False,
        skip_empty: bool = False,
        canvas_size: int = 1920,
        min_size: int = 100,
        output: Optional[str] = None,
    ):
        """Generate a sprite image for the query and its matching images in this Document object.

        An image sprite is a collection of images put into a single image. Query image is on the left
        followed by matching images. The Document object should contain matches.

        :param top_k: the number of top matching documents to show in the sprite.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param inv_normalize: If set to True, inverse the normalization of a float32 image :attr:`.tensor` into a uint8
            image :attr:`.tensor` inplace.
        :param skip_empty: skip matches which has no .uri or .tensor.
        :param canvas_size: the width of the canvas
        :param min_size: the minimum size of the image
        :param output: Optional path to store the visualization. If not given, show in UI
        """
        if not self or not self.matches:
            raise ValueError(f'{self!r} is empty or has no matches')

        if not self.uri and self.tensor is None:
            raise ValueError(
                f'Document has neither `uri` nor `tensor`, cannot be plotted'
            )

        if top_k <= 0:
            raise ValueError(f'`limit` must be larger than 0, receiving {top_k}')

        import matplotlib.pyplot as plt

        img_per_row = top_k + 2
        if top_k > len(self.matches):
            img_per_row = len(self.matches) + 2

        img_size = int((canvas_size - 50) / img_per_row)
        if img_size < min_size:
            # image is too small, recompute the image size and canvas size
            img_size = min_size
            canvas_size = img_per_row * img_size + 50

        try:
            _d = copy.deepcopy(self)
            if _d.content_type != 'tensor':
                _d.load_uri_to_image_tensor()  # the channel axis is -1

            if inv_normalize:
                # inverse normalise to uint8 and set the channel axis to -1
                _d.set_image_tensor_inv_normalization(channel_axis)

            _d.set_image_tensor_channel_axis(channel_axis, -1)

            # Maintain the aspect ratio keeping the width fixed
            h, w, _ = _d.tensor.shape
            img_h, img_w = int(h * (img_size / float(w))), img_size

            sprite_img = np.ones([img_h + 20, canvas_size, 3], dtype='uint8')

            _d.set_image_tensor_shape(shape=(img_h, img_w))

            sprite_img[10 : img_h + 10, 10 : 10 + img_w] = _d.tensor
            pos = canvas_size // img_per_row

            for col_id, d in enumerate(self.matches, start=2):
                if not d.uri and d.tensor is None:
                    if skip_empty:
                        continue
                    else:
                        raise ValueError(
                            f'Document match has neither `uri` nor `tensor`, cannot be plotted'
                        )
                _d = copy.deepcopy(d)
                if _d.content_type != 'tensor':
                    _d.load_uri_to_image_tensor()

                if inv_normalize:
                    _d.set_image_tensor_inv_normalization(channel_axis=channel_axis)

                _d.set_image_tensor_channel_axis(
                    channel_axis, -1
                ).set_image_tensor_shape(shape=(img_h, img_w))

                # paste it on the main canvas
                sprite_img[
                    10 : img_h + 10,
                    (col_id * pos) : ((col_id * pos) + img_w),
                ] = _d.tensor

                col_id += 1
                if col_id >= img_per_row:
                    break
        except Exception as ex:
            raise ValueError('Bad image tensor. Try different `channel_axis`') from ex
        from PIL import Image

        im = Image.fromarray(sprite_img)

        if output:
            with open(output, 'wb') as fp:
                im.save(fp)
        else:
            plt.figure(figsize=(img_per_row, 2))
            plt.gca().set_axis_off()
            plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
            plt.margins(0, 0)
            plt.gca().xaxis.set_major_locator(plt.NullLocator())
            plt.gca().yaxis.set_major_locator(plt.NullLocator())
            plt.imshow(im, interpolation="none")
            plt.show()


def _convert_display_uri(uri, mime_type):
    import urllib
    from docarray.document.mixins.helper import _to_datauri, _uri_to_blob

    scheme = urllib.parse.urlparse(uri).scheme

    if scheme not in ['data', 'http', 'https']:
        blob = _uri_to_blob(uri)
        return _to_datauri(mime_type, blob)
    return uri


def _html5_image(uri):
    from IPython.display import display
    from IPython.core.display import HTML  # noqa

    src = f'''
    <body>
    <image src="{uri}" height="200px">
    </body>
    '''
    display(HTML(src))  # noqa


def _html5_video_player(uri):
    from IPython.display import display
    from IPython.core.display import HTML  # noqa

    src = f'''
    <body>
    <video width="320" height="240" autoplay muted controls>
    <source src="{uri}">
    Your browser does not support the video tag.
    </video>
    </body>
    '''
    display(HTML(src))  # noqa


def _html5_audio_player(uri):
    from IPython.display import display
    from IPython.core.display import HTML  # noqa

    src = f'''
    <body>
    <audio controls="controls" style="width:320px" >
      <source src="{uri}"/>
      Your browser does not support the audio element.
    </audio>
    </body>
    '''
    display(HTML(src))  # noqa



from docarray.array.document import DocumentArray
from docarray.array.storage.qdrant import StorageMixins, QdrantConfig

__all__ = ['DocumentArrayQdrant', 'QdrantConfig']


class DocumentArrayQdrant(StorageMixins, DocumentArray):
    """
    DocumentArray that stores Documents in a `Qdrant <https://weaviate.io/>`_ vector search engine.

    .. note::
        This DocumentArray requires `qdrant-client`. You can install it via `pip install "docarray[qdrant]"`.

        To use Qdrant as storage backend, a Qdrant service needs to be running on your machine.

    With this implementation, :meth:`match` and :meth:`find` perform fast (approximate) vector search.
    Additionally, search with filters is supported.

    Example usage:

    .. code-block:: python

        from docarray import DocumentArray

        # connect to running Qdrant service with default configuration (address: http://localhost:6333)
        da = DocumentArray(storage='qdrant', config={'n_dim': 10})

        # connect to a previously persisted DocumentArrayWeaviate by specifying collection_name, host, and port
        da = DocumentArray(
            storage='qdrant',
            config={
                'collection_name': 'persisted',
                'host': 'localhost',
                'port': '6333',
                'n_dim': 10,
            },
        )


    .. seealso::
        For further details, see our :ref:`user guide <qdrant>`.
    """

    def __new__(cls, *args, **kwargs):
        """``__new__`` method for :class:`DocumentArrayQdrant`

        :param *args: list of args to instantiate the object
        :param **kwargs: dict of args to instantiate the object
        :return: the instantiated :class:`DocumentArrayQdrant` object
        """
        return super().__new__(cls)



from abc import ABC

from docarray.array.storage.annlite.backend import BackendMixin, AnnliteConfig
from docarray.array.storage.annlite.find import FindMixin
from docarray.array.storage.annlite.getsetdel import GetSetDelMixin
from docarray.array.storage.annlite.seqlike import SequenceLikeMixin

__all__ = ['StorageMixins', 'AnnliteConfig']


class StorageMixins(FindMixin, BackendMixin, GetSetDelMixin, SequenceLikeMixin, ABC):
    ...



from typing import (
    TYPE_CHECKING,
    TypeVar,
    Sequence,
    List,
    Dict,
    Optional,
    Union,
)

import numpy as np

from docarray import Document, DocumentArray
from docarray.math import ndarray
from docarray.math.helper import EPSILON
from docarray.math.ndarray import to_numpy_array
from docarray.score import NamedScore

if TYPE_CHECKING:
    import tensorflow
    import torch

    WeaviateArrayType = TypeVar(
        'WeaviateArrayType',
        np.ndarray,
        tensorflow.Tensor,
        torch.Tensor,
        Sequence[float],
    )


class FindMixin:
    def _find_similar_vectors(
        self,
        query: 'WeaviateArrayType',
        limit=10,
        filter: Optional[Dict] = None,
        additional: Optional[List] = None,
        sort: Optional[Union[Dict, List]] = None,
        query_params: Optional[Dict] = None,
    ):
        """Returns a subset of documents by the given vector.

        :param query: input supported to be stored in Weaviate. This includes any from the list '[np.ndarray, tensorflow.Tensor, torch.Tensor, Sequence[float]]'
        :param limit: number of retrieved items
        :param filter: the input filter to apply in each stored document
        :param additional: Optional Weaviate flags for meta data
        :param sort: sort parameters performed on matches performed on results
        :param query_params: additional parameters applied to the query outside of the where clause
        :return: a `DocumentArray` containing the `Document` objects that verify the filter.
        """
        query = to_numpy_array(query)
        is_all_zero = np.all(query == 0)
        if is_all_zero:
            query = query + EPSILON

        query_dict = {'vector': query}

        if query_params:
            query_dict.update(query_params)

        _additional = ['id', 'certainty']
        if additional:
            _additional = _additional + additional

        query_builder = (
            self._client.query.get(self._class_name, '_serialized')
            .with_additional(_additional)
            .with_limit(limit)
            .with_near_vector(query_dict)
        )

        if filter:
            query_builder = query_builder.with_where(filter)

        if sort:
            query_builder = query_builder.with_sort(sort)

        results = query_builder.do()

        docs = []
        if 'errors' in results:
            errors = '\n'.join(map(lambda error: error['message'], results['errors']))
            raise ValueError(
                f'find failed, please check your filter query. Errors: \n{errors}'
            )

        found_results = results.get('data', {}).get('Get', {}).get(self._class_name, [])

        # The serialized document is stored in results['data']['Get'][self._class_name]
        for result in found_results:
            doc = Document.from_base64(result['_serialized'], **self._serialize_config)
            certainty = result['_additional']['certainty']

            doc.scores['weaviate_certainty'] = NamedScore(value=certainty)

            if certainty is None:
                doc.scores['cosine_similarity'] = NamedScore(value=None)
            else:
                doc.scores['cosine_similarity'] = NamedScore(value=2 * certainty - 1)

            doc.tags['wid'] = result['_additional']['id']

            if additional:
                for add in additional:
                    doc.tags[f'{add}'] = result['_additional'][add]

            docs.append(doc)

        return DocumentArray(docs)

    def _filter(
        self,
        filter: Dict,
        limit: Optional[Union[int, float]] = 20,
        additional: Optional[List] = None,
        sort: Optional[Union[Dict, List]] = None,
    ) -> 'DocumentArray':
        """Returns a subset of documents by filtering by the given filter (Weaviate `where` filter).

        :param filter: the input filter to apply in each stored document
        :param limit: number of retrieved items
        :param additional: Optional Weaviate flags for meta data
        :param sort: sort parameters performed on matches performed on results
        :return: a `DocumentArray` containing the `Document` objects that verify the filter.
        """
        if not filter:
            return self

        _additional = ['id']
        if additional:
            _additional = _additional + additional

        query_builder = (
            self._client.query.get(self._class_name, '_serialized')
            .with_additional(_additional)
            .with_where(filter)
            .with_limit(limit)
        )

        if sort:
            query_builder = query_builder.with_sort(sort)

        results = query_builder.do()

        docs = []
        if 'errors' in results:
            errors = '\n'.join(map(lambda error: error['message'], results['errors']))
            raise ValueError(
                f'filter failed, please check your filter query. Errors: \n{errors}'
            )

        found_results = results.get('data', {}).get('Get', {}).get(self._class_name, [])

        # The serialized document is stored in results['data']['Get'][self._class_name]
        for result in found_results:
            doc = Document.from_base64(result['_serialized'], **self._serialize_config)

            doc.tags['wid'] = result['_additional']['id']

            if additional:
                for add in additional:
                    doc.tags[f'{add}'] = result['_additional'][add]

            docs.append(doc)

        return DocumentArray(docs)

    def _find(
        self,
        query: 'WeaviateArrayType',
        limit: int = 10,
        filter: Optional[Dict] = None,
        additional: Optional[List] = None,
        sort: Optional[Union[Dict, List]] = None,
        query_params: Optional[Dict] = None,
        **kwargs,
    ) -> List['DocumentArray']:
        """Returns approximate nearest neighbors given a batch of input queries.
        :param query: input supported to be stored in Weaviate. This includes any from the list '[np.ndarray, tensorflow.Tensor, torch.Tensor, Sequence[float]]'
        :param limit: number of retrieved items
        :param filter: filter query used for pre-filtering
        :param additional: Optional Weaviate flags for meta data
        :param sort: sort parameters performed on matches performed on results
        :param query_params: additional parameters applied to the query outside of the where clause

        :return: DocumentArray containing the closest documents to the query if it is a single query, otherwise a list of DocumentArrays containing
           the closest Document objects for each of the queries in `query`.

        Note: Weaviate returns `certainty` values. To get cosine similarities one needs to use `cosine_sim = 2*certainty - 1` as explained here:
                  https://weaviate.io/developers/weaviate/current/more-resources/faq.html#q-how-do-i-get-the-cosine-similarity-from-weaviates-certainty
        """

        num_rows, _ = ndarray.get_array_rows(query)

        if num_rows == 1:
            return [
                self._find_similar_vectors(
                    query,
                    limit=limit,
                    additional=additional,
                    filter=filter,
                    sort=sort,
                    query_params=query_params,
                )
            ]
        else:
            closest_docs = []
            for q in query:
                da = self._find_similar_vectors(
                    q,
                    limit=limit,
                    additional=additional,
                    filter=filter,
                    sort=sort,
                    query_params=query_params,
                )
                closest_docs.append(da)
            return closest_docs



import sys
from os import path

from setuptools import find_packages
from setuptools import setup

if sys.version_info < (3, 7, 0):
    raise OSError(f'DocArray requires Python >=3.7, but yours is {sys.version}')

try:
    pkg_name = 'docarray'
    libinfo_py = path.join(pkg_name, '__init__.py')
    libinfo_content = open(libinfo_py, 'r', encoding='utf8').readlines()
    version_line = [l.strip() for l in libinfo_content if l.startswith('__version__')][
        0
    ]
    exec(version_line)  # gives __version__
except FileNotFoundError:
    __version__ = '0.0.0'

try:
    with open('README.md', encoding='utf8') as fp:
        _long_description = fp.read()
except FileNotFoundError:
    _long_description = ''

setup(
    name=pkg_name,
    packages=find_packages(),
    version=__version__,
    include_package_data=True,
    description='The data structure for unstructured data',
    author='Jina AI',
    author_email='hello@jina.ai',
    license='Apache 2.0',
    url='https://github.com/jina-ai/docarray',
    download_url='https://github.com/jina-ai/docarray/tags',
    long_description=_long_description,
    long_description_content_type='text/markdown',
    zip_safe=False,
    setup_requires=['setuptools>=18.0', 'wheel'],
    install_requires=['numpy', 'rich>=12.0.0'],
    extras_require={
        # req usage, please see https://docarray.jina.ai/#install
        'common': [
            'protobuf>=3.13.0,<=3.20.1',
            'lz4',
            'requests',
            'matplotlib',
            'Pillow',
            'fastapi',
            'uvicorn',
            'jina-hubble-sdk>=0.11.0',
        ],
        'full': [
            'protobuf>=3.13.0,<=3.20.1',
            'lz4',
            'requests',
            'matplotlib',
            'Pillow',
            'trimesh',
            'scipy',
            'jina-hubble-sdk>=0.10.0',
            'av',
            'fastapi',
            'uvicorn',
            'strawberry-graphql',
            'weaviate-client~=3.3.0',
            'annlite>=0.3.2',
            'qdrant-client~=0.7.3',
            'elasticsearch>=8.2.0',
        ],
        'qdrant': [
            'qdrant-client~=0.7.3',
        ],
        'annlite': [
            'annlite>=0.3.2',
        ],
        'weaviate': [
            'weaviate-client~=3.3.0',
        ],
        'elasticsearch': [
            'elasticsearch>=8.2.0',
        ],
        'test': [
            'pytest',
            'pytest-timeout',
            'pytest-mock',
            'pytest-cov',
            'pytest-repeat',
            'pytest-reraise',
            'mock',
            'pytest-custom_exit_code',
            'black==22.3.0',
            'tensorflow==2.7.0',
            'paddlepaddle==2.2.0',
            'torch==1.9.0',
            'torchvision==0.10.0',
            'datasets',
            'onnx',
            'onnxruntime',
            'jupyterlab',
            'transformers>=4.16.2',
            'weaviate-client~=3.3.0',
            'annlite>=0.3.2',
            'elasticsearch>=8.2.0',
            'jina',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Unix Shell',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Multimedia :: Video',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    project_urls={
        'Documentation': 'https://docarray.jina.ai',
        'Source': 'https://github.com/jina-ai/docarray/',
        'Tracker': 'https://github.com/jina-ai/docarray/issues',
    },
    keywords='docarray deep-learning data-structures cross-modal multi-modal unstructured-data nested-data neural-search',
)



from operator import itemgetter
from typing import Sequence, Iterable

from docarray.array.storage.base.getsetdel import BaseGetSetDelMixin
from docarray.array.storage.base.helper import Offset2ID
from docarray import Document


class GetSetDelMixin(BaseGetSetDelMixin):
    """Implement required and derived functions that power `getitem`, `setitem`, `delitem`"""

    # essential methods start

    def _del_doc_by_id(self, _id: str):
        self._sql(f'DELETE FROM {self._table_name} WHERE doc_id=?', (_id,))
        self._save_offset2ids()
        self._commit()

    def _set_doc_by_id(self, _id: str, value: 'Document'):
        self._sql(
            f'UPDATE {self._table_name} SET serialized_value=?, doc_id=? WHERE doc_id=?',
            (value, value.id, _id),
        )
        self._commit()

    def _get_doc_by_id(self, id: str) -> 'Document':
        r = self._sql(
            f'SELECT serialized_value FROM {self._table_name} WHERE doc_id = ?', (id,)
        )
        res = r.fetchone()
        if res is None:
            raise KeyError(f'Can not find Document with id=`{id}`')
        return res[0]

    # essentials end here

    # now start the optimized bulk methods
    def _get_docs_by_offsets(self, offsets: Sequence[int]) -> Iterable['Document']:
        ids = [self._offset2ids.get_id(offset) for offset in offsets]
        return self._get_docs_by_ids(ids)

    def _clear_storage(self):
        self._sql(f'DELETE FROM {self._table_name}')
        self._commit()

    def _del_docs_by_ids(self, ids: str) -> Iterable['Document']:
        self._sql(
            f"DELETE FROM {self._table_name} WHERE doc_id in ({','.join(['?'] * len(ids))})",
            ids,
        )
        self._save_offset2ids()
        self._commit()

    def _load_offset2ids(self):
        r = self._sql(
            f"SELECT doc_id FROM {self._table_name} ORDER BY item_order",
        )
        self._offset2ids = Offset2ID(list(map(itemgetter(0), r)))

    def _save_offset2ids(self):
        for offset, doc_id in enumerate(self._offset2ids):
            self._sql(
                f"""
                    UPDATE {self._table_name} SET item_order = ? WHERE {self._table_name}.doc_id = ?
                """,
                (offset, doc_id),
            )
        self._commit()

    def _del_docs(self, ids):
        super()._del_docs(ids)
        self._save_offset2ids()

    def _del_doc_by_offset(self, offset: int):
        super()._del_doc_by_offset(offset)
        self._save_offset2ids()



from abc import ABC

from docarray.array.mixins.content import ContentPropertyMixin
from docarray.array.mixins.delitem import DelItemMixin
from docarray.array.mixins.embed import EmbedMixin
from docarray.array.mixins.empty import EmptyMixin
from docarray.array.mixins.evaluation import EvaluationMixin
from docarray.array.mixins.find import FindMixin
from docarray.array.mixins.getattr import GetAttributeMixin
from docarray.array.mixins.getitem import GetItemMixin
from docarray.array.mixins.group import GroupMixin
from docarray.array.mixins.io.binary import BinaryIOMixin
from docarray.array.mixins.io.common import CommonIOMixin
from docarray.array.mixins.io.csv import CsvIOMixin
from docarray.array.mixins.io.dataframe import DataframeIOMixin
from docarray.array.mixins.io.from_gen import FromGeneratorMixin
from docarray.array.mixins.io.json import JsonIOMixin
from docarray.array.mixins.io.pushpull import PushPullMixin
from docarray.array.mixins.match import MatchMixin
from docarray.array.mixins.parallel import ParallelMixin
from docarray.array.mixins.plot import PlotMixin
from docarray.array.mixins.post import PostMixin
from docarray.array.mixins.pydantic import PydanticMixin
from docarray.array.mixins.reduce import ReduceMixin
from docarray.array.mixins.sample import SampleMixin
from docarray.array.mixins.setitem import SetItemMixin
from docarray.array.mixins.strawberry import StrawberryMixin
from docarray.array.mixins.text import TextToolsMixin
from docarray.array.mixins.traverse import TraverseMixin
from docarray.array.mixins.dataloader import DataLoaderMixin


class AllMixins(
    GetAttributeMixin,
    GetItemMixin,
    SetItemMixin,
    DelItemMixin,
    ContentPropertyMixin,
    PydanticMixin,
    StrawberryMixin,
    GroupMixin,
    EmptyMixin,
    CsvIOMixin,
    JsonIOMixin,
    BinaryIOMixin,
    CommonIOMixin,
    EmbedMixin,
    PushPullMixin,
    FromGeneratorMixin,
    FindMixin,
    MatchMixin,
    TraverseMixin,
    PlotMixin,
    SampleMixin,
    PostMixin,
    TextToolsMixin,
    EvaluationMixin,
    ReduceMixin,
    ParallelMixin,
    DataframeIOMixin,
    DataLoaderMixin,
    ABC,
):
    """All plugins that can be used in :class:`DocumentArray`."""

    ...



from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    TypeVar,
    Sequence,
    List,
    Dict,
    Optional,
)

from qdrant_client.http.models.models import Distance

from docarray import Document, DocumentArray
from docarray.math import ndarray
from docarray.score import NamedScore

if TYPE_CHECKING:
    import tensorflow
    import torch
    import numpy as np
    from qdrant_client import QdrantClient

    QdrantArrayType = TypeVar(
        'QdrantArrayType',
        np.ndarray,
        tensorflow.Tensor,
        torch.Tensor,
        Sequence[float],
    )


class FindMixin:
    @property
    @abstractmethod
    def client(self) -> 'QdrantClient':
        raise NotImplementedError()

    @property
    @abstractmethod
    def collection_name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def serialize_config(self) -> dict:
        raise NotImplementedError()

    @property
    @abstractmethod
    def distance(self) -> 'Distance':
        raise NotImplementedError()

    def _find_similar_vectors(
        self, q: 'QdrantArrayType', limit: int = 10, filter: Optional[Dict] = None
    ):
        query_vector = self._map_embedding(q)

        search_result = self.client.search(
            self.collection_name,
            query_vector=query_vector,
            query_filter=filter,
            search_params=None,
            top=limit,
            append_payload=['_serialized'],
        )

        docs = []

        for hit in search_result:
            doc = Document.from_base64(
                hit.payload['_serialized'], **self.serialize_config
            )
            doc.scores[f'{self.distance.lower()}_similarity'] = NamedScore(
                value=hit.score
            )
            docs.append(doc)

        return DocumentArray(docs)

    def _find(
        self,
        query: 'QdrantArrayType',
        limit: int = 10,
        filter: Optional[Dict] = None,
        **kwargs,
    ) -> List['DocumentArray']:
        """Returns approximate nearest neighbors given a batch of input queries.
        :param query: input supported to be used in Qdrant.
        :param limit: number of retrieved items
        :param filter: filter query used for pre-filtering


        :return: a list of DocumentArrays containing the closest Document objects for each of the queries in `query`.
        """

        num_rows, _ = ndarray.get_array_rows(query)

        if num_rows == 1:
            return [self._find_similar_vectors(query, limit=limit, filter=filter)]
        else:
            closest_docs = []
            for q in query:
                da = self._find_similar_vectors(q, limit=limit, filter=filter)
                closest_docs.append(da)
            return closest_docs
